# -*- coding: utf-8 -*-

import os
import numpy as np
import pandas as pd
from datetime import datetime, timedelta
from hbshare.fe.common.util.verifier import verify_type
from hbshare.fe.common.util.exception import PreprocessError
from hbshare.fe.common.util.logger import logger
from hbshare.fe.xwq.analysis.orm.fedb import FEDB
from hbshare.fe.xwq.analysis.orm.hbdb import HBDB


class IndexConstruction:
    def __init__(self, index_name, date, data_path, update_all):
        """
        :param index_name: 指数名称
        :param start_date: 归因的起始时间
        :param end_date: 归因的结束时间
        :param frequency: 归因频率
        """
        self.index_name = index_name
        self.index_symbol_dic = {'主动股票型基金指数': 'zdgj', '偏股混合型基金指数': 'pgjj'}
        self.index_symbol = self.index_symbol_dic[self.index_name]
        self.index_fund_type_dic = {'主动股票型基金指数': ['13', '37', '35'], '偏股混合型基金指数': ['37']}
        self.fund_type = self.index_fund_type_dic[self.index_name]
        self.start_date = '20071231'
        self.end_date = date
        self.data_path = data_path
        self.update_all = update_all
        self.exist_limit = 90
        self._verify_input_param()
        self._load_data()
        self._init_data()

    def _verify_input_param(self):
        verify_type(self.index_name, 'index_name', str)
        verify_type(self.start_date, 'start_date', str)
        verify_type(self.end_date, 'end_date', str)
        verify_type(self.data_path, 'data_path', str)

    def preload_fund_cumret(self, start_date, end_date, data_path):
        if not os.path.exists(data_path):
            os.makedirs(data_path)
        fund_cumret_path = data_path + 'fund_cumret.hdf'
        if os.path.isfile(fund_cumret_path):
            existed_fund_cumret = pd.read_hdf(fund_cumret_path, key='table')
            max_date = max(existed_fund_cumret['JZRQ'])
            start_date = max(str(max_date), start_date)
        else:
            existed_fund_cumret = pd.DataFrame()
        trade_df = self.trade_df[(self.trade_df['TRADE_DATE'] > start_date) & (self.trade_df['TRADE_DATE'] <= end_date)]
        data = []
        for idx, td in enumerate(trade_df['TRADE_DATE'].unique().tolist()):
            fund_cumret_date = HBDB().read_fund_cumret_given_date(td)
            data.append(fund_cumret_date)
            print('[PreloadFundCumret][{0}/{1}]'.format(idx, len(trade_df['TRADE_DATE'])))
        fund_cumret = pd.concat([existed_fund_cumret] + data, ignore_index=True)
        # fund_cumret = fund_cumret.drop_duplicates(['JJDM', 'JZRQ'])
        fund_cumret.to_hdf(fund_cumret_path, key='table', mode='w')
        return

    def _load_data(self):
        self.calendar_df = HBDB().read_cal((datetime.strptime(self.start_date, '%Y%m%d') - timedelta(365)).strftime('%Y%m%d'), self.end_date)
        self.calendar_df = self.calendar_df.rename(columns={'JYRQ': 'CALENDAR_DATE', 'SFJJ': 'IS_OPEN', 'SFZM': 'IS_WEEK_END', 'SFYM': 'IS_MONTH_END'})
        self.calendar_df['CALENDAR_DATE'] = self.calendar_df['CALENDAR_DATE'].astype(str)
        self.calendar_df = self.calendar_df.sort_values('CALENDAR_DATE')
        self.calendar_df['IS_OPEN'] = self.calendar_df['IS_OPEN'].astype(int).replace({0: 1, 1: 0})
        self.calendar_df['YEAR_MONTH'] = self.calendar_df['CALENDAR_DATE'].apply(lambda x: x[:6])
        self.calendar_df['MONTH_DAY'] = self.calendar_df['CALENDAR_DATE'].apply(lambda x: x[4:])
        self.report_df = self.calendar_df.drop_duplicates('YEAR_MONTH', keep='last').rename(columns={'CALENDAR_DATE': 'REPORT_DATE'})
        self.report_df = self.report_df[self.report_df['MONTH_DAY'].isin(['0331', '0630', '0930', '1231'])]
        self.trade_df = self.calendar_df[self.calendar_df['IS_OPEN'] == 1].rename(columns={'CALENDAR_DATE': 'TRADE_DATE'})
        self.first_trade_date = self.trade_df[self.trade_df['TRADE_DATE'] <= self.start_date]['TRADE_DATE'].iloc[-1]
        self.trade_df = self.trade_df[(self.trade_df['TRADE_DATE'] >= self.start_date) & (self.trade_df['TRADE_DATE'] <= self.end_date)]

        self.fund_info = HBDB().read_fund_info()
        self.fund_info = self.fund_info.rename(columns={'jjdm': 'FUND_CODE', 'jjmc': 'FUND_FULL_NAME', 'jjjc': 'FUND_SHORT_NAME', 'clrq': 'ESTABLISH_DATE', 'zzrq': 'EXPIRE_DATE', 'cpfl': 'PRODUCT_TYPE', 'kffb': 'OPEN_CLOSE', 'jjfl': 'FUND_TYPE_1ST', 'ejfl': 'FUND_TYPE_2ND'})
        self.fund_info = self.fund_info.dropna(subset=['ESTABLISH_DATE'])
        self.fund_info['EXPIRE_DATE'] = self.fund_info['EXPIRE_DATE'].fillna(20990101)
        self.fund_info['ESTABLISH_DATE'] = self.fund_info['ESTABLISH_DATE'].astype(int).astype(str)
        self.fund_info['EXPIRE_DATE'] = self.fund_info['EXPIRE_DATE'].astype(int).astype(str)
        self.fund_info = self.fund_info[(self.fund_info['PRODUCT_TYPE'] == '2') & (self.fund_info['OPEN_CLOSE'] == '0')]
        self.fund_info = self.fund_info[self.fund_info['FUND_TYPE_2ND'].isin(self.fund_type)]

        self.preload_fund_cumret(self.start_date, self.end_date, self.data_path)
        self.fund_cumret = pd.read_hdf(self.data_path + 'fund_cumret.hdf', key='table')
        self.fund_cumret = self.fund_cumret.rename(columns={'JJDM': 'FUND_CODE', 'JZRQ': 'TRADE_DATE', 'HBCL': 'CUM_RET'})
        self.fund_cumret['TRADE_DATE'] = self.fund_cumret['TRADE_DATE'].astype(str)
        self.first_fund_cumret = HBDB().read_fund_cumret_given_date(self.first_trade_date)
        self.first_fund_cumret = self.first_fund_cumret.rename(columns={'JJDM': 'FUND_CODE', 'JZRQ': 'TRADE_DATE', 'HBCL': 'CUM_RET'})
        self.first_fund_cumret['TRADE_DATE'] = self.first_fund_cumret['TRADE_DATE'].astype(str)
        self.fund_cumret = pd.concat([self.first_fund_cumret, self.fund_cumret])
        self.fund_cumret = self.fund_cumret.pivot(index='TRADE_DATE', columns='FUND_CODE', values='CUM_RET')
        self.fund_cumret = self.fund_cumret.sort_index()
        self.fund_nav = 0.01 * self.fund_cumret + 1

        fund_aum_list = []
        for idx, td in enumerate(self.report_df['REPORT_DATE'].unique().tolist()):
            fund_aum_date = HBDB().read_fund_scale_given_date(td)
            if len(fund_aum_date) == 0:
                fund_aum_date = pd.DataFrame(columns=['ZCJZ', 'BBLB1', 'JSRQ', 'JJDM', 'GGRQ', 'ROW_ID'])
            fund_aum_date = fund_aum_date[fund_aum_date['BBLB1'] == 13]
            fund_aum_date = fund_aum_date.sort_values(['JJDM', 'JSRQ', 'GGRQ']).drop_duplicates(['JJDM', 'JSRQ'], keep='last')
            fund_aum_list.append(fund_aum_date)
            print('[PreloadFundAum][{0}/{1}]'.format(idx, len(self.report_df['REPORT_DATE'])))
        fund_aum = pd.concat(fund_aum_list, ignore_index=True)
        # fund_aum = fund_aum.drop_duplicates(['JJDM', 'JSRQ'])
        fund_aum.to_hdf(data_path + 'fund_aum.hdf', key='table', mode='w')
        self.fund_aum = pd.read_hdf(self.data_path + 'fund_aum.hdf', key='table')
        self.fund_aum = self.fund_aum.rename(columns={'JJDM': 'FUND_CODE', 'JSRQ': 'REPORT_DATE', 'ZCJZ': 'AUM'})
        self.fund_aum['REPORT_DATE'] = self.fund_aum['REPORT_DATE'].astype(str)
        self.fund_aum = self.fund_aum.pivot(index='REPORT_DATE', columns='FUND_CODE', values='AUM')
        self.fund_aum = self.fund_aum.sort_index()

        fund_gptzzjb_list = []
        for idx, td in enumerate(self.report_df['REPORT_DATE'].unique().tolist()):
            fund_gptzzjb_date = HBDB().read_fund_gptzzjb_given_date(td)
            if len(fund_gptzzjb_date) == 0:
                fund_gptzzjb_date = pd.DataFrame(columns=['jjdm', 'jsrq', 'gptzzjb'])
            fund_gptzzjb_list.append(fund_gptzzjb_date)
            print('[PreloadFundStockRatio][{0}/{1}]'.format(idx, len(self.report_df['REPORT_DATE'])))
        fund_gptzzjb = pd.concat(fund_gptzzjb_list, ignore_index=True)
        fund_gptzzjb.to_hdf(data_path + 'fund_gptzzjb.hdf', key='table', mode='w')
        self.fund_gptzzjb = pd.read_hdf(self.data_path + 'fund_gptzzjb.hdf', key='table')
        self.fund_gptzzjb = self.fund_gptzzjb.rename(columns={'jjdm': 'FUND_CODE', 'jsrq': 'REPORT_DATE', 'gptzzjb': 'STOCK_RATIO'})
        self.fund_gptzzjb['REPORT_DATE'] = self.fund_gptzzjb['REPORT_DATE'].astype(str)
        self.fund_gptzzjb = self.fund_gptzzjb.pivot(index='REPORT_DATE', columns='FUND_CODE', values='STOCK_RATIO')
        self.fund_gptzzjb = self.fund_gptzzjb.sort_index()

    def _init_data(self):
        date_list = sorted(list(set([self.first_trade_date] + self.trade_df['TRADE_DATE'].unique().tolist())))
        if not self.update_all:
            latest_date = FEDB().read_fund_index_latest_date()
            date_list = [date for date in date_list if date >= latest_date]
        if not (self.fund_nav.empty or self.fund_aum.empty):
                self.fund_nav = self.fund_nav.reindex(date_list).interpolate().sort_index()
                self.fund_aum = self.calendar_df[['CALENDAR_DATE']].set_index('CALENDAR_DATE').sort_index().merge(self.fund_aum, left_index=True, right_index=True, how='left')
                self.fund_aum = self.fund_aum.fillna(method='ffill').reindex(date_list)
                self.fund_gptzzjb = self.fund_gptzzjb.rolling(window=3, min_periods=1).min()
                self.fund_gptzzjb = self.fund_gptzzjb.fillna(0.0)
                self.fund_gptzzjb = self.calendar_df[['CALENDAR_DATE']].set_index('CALENDAR_DATE').sort_index().merge(self.fund_gptzzjb, left_index=True, right_index=True, how='left')
                self.fund_gptzzjb = self.fund_gptzzjb.fillna(method='ffill').reindex(date_list)
                assert (self.fund_nav.shape[0] == self.fund_aum.shape[0])
                assert (self.fund_nav.shape[0] == self.fund_gptzzjb.shape[0])

                self.fund_info['INTO_DATE'] = self.fund_info['ESTABLISH_DATE'].apply(lambda x: (datetime.strptime(x, '%Y%m%d') + timedelta(self.exist_limit)).strftime('%Y%m%d'))
                self.fund_info['OUT_DATE'] = self.fund_info['EXPIRE_DATE']
                fund_nav_unstack = self.fund_nav.unstack().reset_index().rename(columns={0: 'NAV_ADJ'})
                fund_nav_unstack = fund_nav_unstack.merge(self.fund_info[['FUND_CODE', 'INTO_DATE', 'OUT_DATE']], on=['FUND_CODE'], how='left')
                fund_nav_unstack['INTO_DATE'] = fund_nav_unstack['INTO_DATE'].fillna('20990101')
                fund_nav_unstack['OUT_DATE'] = fund_nav_unstack['OUT_DATE'].fillna('19000101')
                if self.index_name == '主动股票型基金指数':
                    fund_gptzzjb_unstack = self.fund_gptzzjb.unstack().reset_index()
                    fund_gptzzjb_unstack.columns = ['FUND_CODE', 'TRADE_DATE', 'STOCK_RATIO']
                    fund_nav_unstack = fund_nav_unstack.merge(fund_gptzzjb_unstack, on=['FUND_CODE', 'TRADE_DATE'], how='left')
                    fund_nav_unstack = fund_nav_unstack.dropna()
                    fund_nav_unstack = fund_nav_unstack[fund_nav_unstack['STOCK_RATIO'] >= 50.0]
                fund_nav_unstack['INTO_SAMPLE'] = (fund_nav_unstack['INTO_DATE'] <= fund_nav_unstack['TRADE_DATE']) & (fund_nav_unstack['OUT_DATE'] > fund_nav_unstack['TRADE_DATE'])
                fund_nav_unstack['INTO_SAMPLE'] = fund_nav_unstack['INTO_SAMPLE'].astype(int)
                self.into_sample = fund_nav_unstack.pivot(index='TRADE_DATE', columns='FUND_CODE', values='INTO_SAMPLE')
                self.into_sample = self.into_sample.reindex(date_list).fillna(0.0)
                assert (self.fund_nav.shape[0] == self.into_sample.shape[0])
        else:
            msg = "Data empty occurred, check your input"
            logger.error(msg)
            raise PreprocessError(message=msg)

    def get_index(self):
        fund_ret = self.fund_nav / self.fund_nav.shift()
        fund_total_weight = pd.DataFrame((self.fund_aum * self.into_sample).sum(axis=1)).rename(columns={0: 'TOTAL_AUM'})
        fund_total_weight['TOTAL_AUM'] = fund_total_weight['TOTAL_AUM'].replace(0.0, np.nan)
        fund_weight = (self.fund_aum * self.into_sample).merge(fund_total_weight, left_index=True, right_index=True, how='left')
        fund_weight = fund_weight.apply(lambda x: x[:-1] / x[-1], axis=1)
        fund_index_adj = (fund_ret * fund_weight).sum(axis=1)
        if self.update_all:
            fund_index_adj.iloc[0] = 1000
            fund_index_adj = fund_index_adj.rename(index={fund_index_adj.index[0]: self.start_date})
            index = fund_index_adj.cumprod()
            index_df = pd.DataFrame(index).reset_index()
            index_df.columns = ['TRADE_DATE', 'INDEX_POINT']
            index_df['INDEX_SYMBOL'] = self.index_symbol
            index_df['INDEX_NAME'] = self.index_name
            index_df = index_df[['TRADE_DATE', 'INDEX_SYMBOL', 'INDEX_NAME', 'INDEX_POINT']]
            index_df['TRADE_DATE'] = index_df['TRADE_DATE'].astype(str)
            index_df = index_df.sort_values('TRADE_DATE')
            index_df['YEAR'] = index_df['TRADE_DATE'].apply(lambda x: x[:4])
            index_df_year = index_df.drop_duplicates('YEAR', keep='last')
            index_df_year['YEAR'] = index_df_year['YEAR'].apply(lambda x: str(int(x) + 1))
            index_df_year = index_df_year.rename(columns={'INDEX_POINT': 'INDEX_POINT_YEAR'})
            index_df = index_df.merge(index_df_year[['YEAR', 'INDEX_POINT_YEAR']], on=['YEAR'], how='left')
            index_df['涨幅'] = index_df['INDEX_POINT'].pct_change() * 100.0
            index_df['近1月涨幅'] = index_df['INDEX_POINT'].pct_change(20) * 100.0
            index_df['近3月涨幅'] = index_df['INDEX_POINT'].pct_change(60) * 100.0
            index_df['近6月涨幅'] = index_df['INDEX_POINT'].pct_change(120) * 100.0
            index_df['近1年涨幅'] = index_df['INDEX_POINT'].pct_change(250) * 100.0
            index_df['今年以来涨幅'] = (index_df['INDEX_POINT'] / index_df['INDEX_POINT_YEAR'] - 1.0) * 100.0
            index_df['成立以来涨幅'] = (index_df['INDEX_POINT'] / 1000.0 - 1.0) * 100.0
            index_df = index_df.fillna(99999)
            index_df = index_df[['TRADE_DATE', 'INDEX_SYMBOL', 'INDEX_NAME', 'INDEX_POINT', '涨幅', '近1月涨幅', '近3月涨幅', '近6月涨幅', '近1年涨幅', '今年以来涨幅', '成立以来涨幅']]
            FEDB().insert_fund_index(index_df)
        else:
            latest_date = FEDB().read_fund_index_latest_date()
            latest_index = FEDB().read_fund_index_given_date(latest_date)
            latest_index = latest_index[latest_index['INDEX_NAME'] == self.index_name]['INDEX_POINT'].values[0]
            fund_index_adj.iloc[0] = latest_index
            index = fund_index_adj.cumprod()
            index_df = pd.DataFrame(index).reset_index().iloc[1:]
            index_df.columns = ['TRADE_DATE', 'INDEX_POINT']
            index_df['INDEX_SYMBOL'] = self.index_symbol
            index_df['INDEX_NAME'] = self.index_name
            index_df = index_df[['TRADE_DATE', 'INDEX_SYMBOL', 'INDEX_NAME', 'INDEX_POINT']]
            index_df_db = FEDB().read_fund_index_gt_date('20071231')[['TRADE_DATE', 'INDEX_SYMBOL', 'INDEX_NAME', 'INDEX_POINT']]
            index_df_db = index_df_db[index_df_db['INDEX_NAME'] == self.index_name]
            index_df = pd.concat([index_df_db, index_df])
            index_df['TRADE_DATE'] = index_df['TRADE_DATE'].astype(str)
            index_df = index_df.sort_values('TRADE_DATE')
            index_df = index_df.drop_duplicates('TRADE_DATE')
            index_df['YEAR'] = index_df['TRADE_DATE'].apply(lambda x: x[:4])
            index_df_year = index_df.drop_duplicates('YEAR', keep='last')
            index_df_year['YEAR'] = index_df_year['YEAR'].apply(lambda x: str(int(x) + 1))
            index_df_year = index_df_year.rename(columns={'INDEX_POINT': 'INDEX_POINT_YEAR'})
            index_df = index_df.merge(index_df_year[['YEAR', 'INDEX_POINT_YEAR']], on=['YEAR'], how='left')
            index_df['涨幅'] = index_df['INDEX_POINT'].pct_change() * 100.0
            index_df['近1月涨幅'] = index_df['INDEX_POINT'].pct_change(20) * 100.0
            index_df['近3月涨幅'] = index_df['INDEX_POINT'].pct_change(60) * 100.0
            index_df['近6月涨幅'] = index_df['INDEX_POINT'].pct_change(120) * 100.0
            index_df['近1年涨幅'] = index_df['INDEX_POINT'].pct_change(250) * 100.0
            index_df['今年以来涨幅'] = (index_df['INDEX_POINT'] / index_df['INDEX_POINT_YEAR'] - 1.0) * 100.0
            index_df['成立以来涨幅'] = (index_df['INDEX_POINT'] / 1000.0 - 1.0) * 100.0
            index_df = index_df.fillna(99999)
            index_df = index_df[['TRADE_DATE', 'INDEX_SYMBOL', 'INDEX_NAME', 'INDEX_POINT', '涨幅', '近1月涨幅', '近3月涨幅', '近6月涨幅', '近1年涨幅', '今年以来涨幅', '成立以来涨幅']]
            index_df = index_df[index_df['TRADE_DATE'] > latest_date]
            FEDB().insert_fund_index(index_df)
        return


if __name__ == '__main__':
    today = datetime.today().strftime('%Y%m%d')
    calendar_df = HBDB().read_cal((datetime.strptime(today, '%Y%m%d') - timedelta(15)).strftime('%Y%m%d'), today)
    calendar_df = calendar_df.rename(columns={'JYRQ': 'CALENDAR_DATE', 'SFJJ': 'IS_OPEN', 'SFZM': 'IS_WEEK_END', 'SFYM': 'IS_MONTH_END'})
    calendar_df['CALENDAR_DATE'] = calendar_df['CALENDAR_DATE'].astype(str)
    calendar_df = calendar_df.sort_values('CALENDAR_DATE')
    calendar_df['IS_OPEN'] = calendar_df['IS_OPEN'].astype(int).replace({0: 1, 1: 0})
    trade_df = calendar_df[calendar_df['IS_OPEN'] == 1].rename(columns={'CALENDAR_DATE': 'TRADE_DATE'})
    run_date_today = trade_df[trade_df['TRADE_DATE'] <= today]['TRADE_DATE'].iloc[-1]
    run_date_yesterday = trade_df[trade_df['TRADE_DATE'] <= today]['TRADE_DATE'].iloc[-2]
    if today == run_date_today:
        update_all = False
        data_path = 'D:/Git/hbshare/hbshare/fe/xwq/data/index_construction/'
        for index_name in ['主动股票型基金指数', '偏股混合型基金指数']:
            IndexConstruction(index_name, run_date_yesterday, data_path, update_all).get_index()

    # update_all = False
    # index_name = '主动股票型基金指数'  # 主动股票型基金指数，偏股混合型基金指数
    # date = (datetime.today() - timedelta(1)).strftime('%Y%m%d')
    # data_path = 'D:/Git/hbshare/hbshare/fe/xwq/data/index_construction/'
    # IndexConstruction(index_name, date, data_path, update_all).get_index()

    # index_aum = FEDB().read_fund_index_gt_date('20071231')
    # index_aum = index_aum[['TRADE_DATE', 'INDEX_NAME', 'INDEX_POINT']]
    # index_zz = HBDB().read_index_daily_k_given_date_and_indexs('20071231', ['930890', '930950'])
    # index_zz = index_zz[['zqmc', 'jyrq', 'spjg']].rename(columns={'zqmc': 'INDEX_NAME', 'jyrq': 'TRADE_DATE', 'spjg': 'INDEX_POINT'})
    # index = pd.concat([index_aum, index_zz])
    # index['TRADE_DATE'] = index['TRADE_DATE'].astype(str)
    # index = index.pivot(index='TRADE_DATE', columns='INDEX_NAME', values='INDEX_POINT').fillna(1000).sort_index()
    # index = index.rename(columns={'偏股基金': '中证偏股型基金指数', '主动股基': '中证主动股票型基金指数', '偏股混合型基金指数': '基于AUM的偏股混合型基金指数', '主动股票型基金指数': '基于AUM的主动股票型基金指数'})
    # index.index = map(lambda x: datetime.strptime(x, '%Y%m%d'), index.index)
    #
    # import matplotlib.pyplot as plt
    # plt.rcParams['font.sans-serif'] = ['SimHei']
    # plt.rcParams['axes.unicode_minus'] = False
    #
    # fig, ax = plt.subplots()
    # ax.plot(index.index, index['基于AUM的偏股混合型基金指数'].values, label='基于AUM的偏股混合型基金指数', color='#F04950')
    # ax.plot(index.index, index['中证偏股型基金指数'].values, label='中证偏股型基金指数', color='#6268A2')
    # plt.legend(loc=2)
    # plt.xticks(rotation=90)
    # plt.tight_layout()
    # plt.savefig('D:/Git/hbshare/hbshare/fe/xwq/data/index_construction/偏股型基金指数.png')
    #
    # fig, ax = plt.subplots()
    # ax.plot(index.index, index['基于AUM的主动股票型基金指数'].values, label='基于AUM的主动股票型基金指数', color='#F04950')
    # ax.plot(index.index, index['中证主动股票型基金指数'].values, label='中证主动股票型基金指数', color='#6268A2')
    # plt.legend(loc=2)
    # plt.xticks(rotation=90)
    # plt.tight_layout()
    # plt.savefig('D:/Git/hbshare/hbshare/fe/xwq/data/index_construction/主动股票型基金指数.png')



"""
季度资产报告辅助程序
"""
import pandas as pd
from datetime import datetime
import os
import plotly
from plotly.offline import plot as plot_ly
import plotly.graph_objs as go

plotly.offline.init_notebook_mode(connected=True)


class QuarterAssetAllocationReport:
    def __init__(self, start_date, end_date, data_path, file_name):
        self.start_date = start_date
        self.end_date = end_date
        self.data_path = data_path
        self.file_name = file_name
        self._load_data()

    def _load_data(self):
        data_with_header = pd.read_excel(
            os.path.join(self.data_path, r"{}".format(self.file_name)), sheet_name='原始净值')
        data = pd.read_excel(os.path.join(self.data_path, r"{}".format(self.file_name)), sheet_name='原始净值', header=1)
        data['t_date'] = data['t_date'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
        data.index = data['t_date']
        cols = data_with_header.columns.tolist()
        # 月末日期
        df = data[['t_date']].sort_index()
        df['month'] = df['t_date'].apply(lambda x: datetime.strptime(x, '%Y%m%d').month)
        df['next_month'] = df['month'].shift(-1).fillna(0.).astype(int)
        df = df[(df['month'] != df['next_month']) & (df['t_date'] >= self.start_date)]
        trading_day_list = sorted(df['t_date'].tolist())

        from hbshare.quant.Kevin.quant_room.MyUtil.data_loader import get_trading_day_list
        trading_day_list = get_trading_day_list(self.start_date, self.end_date, frequency="week")

        # type
        data_param = dict()
        type_list = [x for x in cols if not x.startswith('Unnamed')]
        for i in range(len(type_list) - 1):
            typ = type_list[i]
            s_index, e_index = cols.index(type_list[i]), cols.index(type_list[i + 1])
            data_slice = data[data.columns[s_index: e_index]]
            if typ in ['量价（500）', '机器学习', '基本面']:
                data_slice['benchmark'] = data['中证500']
            elif typ in ['量价（300）']:
                data_slice['benchmark'] = data['沪深300']
            elif typ in ['量价（1000）']:
                data_slice['benchmark'] = data['中证1000']
            else:
                 pass
            data_slice = data_slice[data_slice.index >= self.start_date].reindex(trading_day_list)
            data_param[type_list[i]] = data_slice
        # 指数
        index_df = data[['沪深300', '中证500', '中证1000']].reindex(trading_day_list)

        self.trading_day_list = trading_day_list
        self.data_param = data_param
        self.index_df = index_df

    def calculate_excess_return(self):
        data_lj = self.data_param['量价（500）']
        data_ml = self.data_param['机器学习']
        data_fd = self.data_param['基本面']
        data_300 = self.data_param['量价（300）']
        data_1000 = self.data_param['量价（1000）']

        return_lj = data_lj.pct_change(limit=1).dropna(axis=0, how='all')
        return_lj['average'] = return_lj[return_lj.columns[:-1]].mean(axis=1)
        return_lj['ave_excess'] = return_lj['average'] - return_lj['benchmark']

        return_fd = data_fd.pct_change(limit=1).dropna(axis=0, how='all')
        return_fd['average'] = return_fd[return_fd.columns[:-1]].mean(axis=1)
        return_fd['ave_excess'] = return_fd['average'] - return_fd['benchmark']

        return_ml = data_ml.pct_change(limit=1).dropna(axis=0, how='all')
        return_ml['average'] = return_ml[return_ml.columns[:-1]].mean(axis=1)
        return_ml['ave_excess'] = return_ml['average'] - return_ml['benchmark']

        return_300 = data_300.pct_change(limit=1).dropna(axis=0, how='all')
        return_300['average'] = return_300[return_300.columns[:-1]].mean(axis=1)
        return_300['ave_excess'] = return_300['average'] - return_300['benchmark']

        return_1000 = data_1000.pct_change(limit=1).dropna(axis=0, how='all')
        return_1000['average'] = return_1000[return_1000.columns[:-1]].mean(axis=1)
        return_1000['ave_excess'] = return_1000['average'] - return_1000['benchmark']

        data_500 = data_lj[data_lj.columns[:-1]].merge(
            data_fd[data_fd.columns[:-1]], left_index=True, right_index=True).merge(
            data_ml, left_index=True, right_index=True)

        return_500 = data_500.pct_change(limit=1).dropna(axis=0, how='all')
        return_500['average'] = return_500[return_500.columns[:-1]].mean(axis=1)
        return_500['ave_excess'] = return_500['average'] - return_500['benchmark']

        compare_df = return_lj['ave_excess'].to_frame('量价类').merge(
            return_fd['ave_excess'].to_frame('基本面'), left_index=True, right_index=True).merge(
            return_ml['ave_excess'].to_frame('机器学习'), left_index=True, right_index=True)

        compare_df2 = return_300['ave_excess'].to_frame('300指增').merge(
            return_500['ave_excess'].to_frame('500指增'), left_index=True, right_index=True).merge(
            return_1000['ave_excess'].to_frame('1000指增'), left_index=True, right_index=True)

        return compare_df, compare_df2

    def calculate_index_return(self):
        index_df = self.index_df.copy()
        return_df = index_df.pct_change().dropna(axis=0, how='all')

        return return_df

    def calculate_market_vol(self):
        # 直接从本地取
        path = 'D:\\kevin\\risk_model_jy\\RiskModel\\data\\common_data\\chg_pct'
        filenames = os.listdir(path)
        filenames = [x for x in filenames if (
                    not x.split('.')[0] < self.start_date and x.split('.')[0] <= self.end_date)]
        date_list = []
        vol_list = []
        for name in filenames:
            date_list.append(name.split('.')[0])
            df = pd.read_csv(os.path.join(path, name)).dropna()
            df = df[(df['dailyReturnReinv'] >= -0.2) & (df['dailyReturnReinv'] <= 0.2)]
            vol_list.append(df['dailyReturnReinv'].std())

        vol_df = pd.DataFrame({"date": date_list, "vol": vol_list})

        return vol_df


def plotly_bar(df, title_text, save_path):
    cols = df.columns.tolist()
    color_list = ['rgb(216, 0, 18)', 'rgb(60, 127, 175)', 'rgb(35, 35, 35)']
    data = []
    for i in range(len(cols)):
        col = cols[i]
        trace = go.Bar(
            x=df.index.tolist(),
            y=df[col],
            name=col,
            marker=dict(color=color_list[i]),
            # width=0.2,
        )
        data.append(trace)

    layout = go.Layout(
        title=dict(text=title_text, x=0.07, y=0.85),
        autosize=False, width=1200, height=500,
        yaxis=dict(tickfont=dict(size=12), tickformat=',.1%', showgrid=True),
        xaxis=dict(showgrid=True),
        legend=dict(orientation="h", x=0.35),
        template='plotly_white'
    )
    fig = go.Figure(data=data, layout=layout)

    plot_ly(fig, filename=save_path, auto_open=False)


def plotly_line(df, day_list, title_text, save_path):
    trace = go.Scatter(
        x=df['date'],
        y=df['vol'],
        mode="lines",
        marker=dict(color='rgb(35, 35, 35)')
    )

    data = [trace]

    tick_vals = [df['date'].tolist().index(x) for x in day_list]
    tick_text = [x[:6] for x in day_list]

    layout = go.Layout(
        title=dict(text=title_text, x=0.07, y=0.85),
        autosize=False, width=1200, height=400,
        yaxis=dict(tickfont=dict(size=12), tickformat=',.1%', showgrid=True),
        xaxis=dict(showgrid=True, tickvals=tick_vals, ticktext=tick_text),
        template='plotly_white'
    )

    fig = go.Figure(data=data, layout=layout)

    plot_ly(fig, filename=save_path, auto_open=False)


if __name__ == '__main__':
    allo_rep = QuarterAssetAllocationReport(
        start_date='20161230', end_date='20220805',
        data_path='D:\\量化产品跟踪\\指数增强', file_name='指增-20220805.xlsx')
    # allo_rep = QuarterAssetAllocationReport(
    #     start_date='20161230', end_date='20220527',
    #     data_path='D:\\量化产品跟踪\\指数增强', file_name='指增-20220527.xlsx')

    excess_df, excess_df2 = allo_rep.calculate_excess_return()

    excess_df['date'] = excess_df.index
    excess_df['date'] = excess_df['date'].apply(lambda x: x[:6])
    excess_df = excess_df.set_index('date')[['机器学习', '基本面', '量价类']]
    plotly_bar(excess_df, "股票量化策略超额表现", "D:\\市场微观结构图\\股票量化策略超额表现-分策略.html")

    excess_df2['date'] = excess_df2.index
    excess_df2['date'] = excess_df2['date'].apply(lambda x: x[:6])
    excess_df2 = excess_df2.set_index('date')[['300指增', '500指增', '1000指增']]
    plotly_bar(excess_df2, "股票量化策略超额表现", "D:\\市场微观结构图\\股票量化策略超额表现-分基准.html")

    index_return_df = allo_rep.calculate_index_return()
    index_return_df['date'] = index_return_df.index
    index_return_df['date'] = index_return_df['date'].apply(lambda x: x[:6])
    index_return_df = index_return_df.set_index('date')
    plotly_bar(index_return_df, "指数表现", "D:\\市场微观结构图\\指数表现.html")

    market_vol = allo_rep.calculate_market_vol()
    plotly_line(market_vol, allo_rep.trading_day_list, "市场横截面波动率", "D:\\市场微观结构图\\市场横截面波动率.html")



#!/usr/bin/python
# coding:utf-8
import pandas as pd
import numpy as np


def xs_weight(data):
    pos_long = sum(data['POS'] == 1)
    pos_short = sum(data['POS'] == -1)
    data['WEIGHT'] = data['POS'].apply(
        lambda x: 0.5 / pos_long if x == 1 else (0.5 / pos_short if x == -1 else 0)
    )
    return data


def ts_weight(data):
    if sum(data['POS'] != 0) > 0:
        data['WEIGHT'] = 1 / sum(data['POS'] != 0)
    else:
        data['WEIGHT'] = 0

    return data


def ts_pos(data, direction=1):
    data['POS'] = data['FACTORVALUE'] / data['FACTORVALUE'].abs() * direction
    data = data[~np.isnan(data['POS'])].reset_index(drop=True)
    return ts_weight(data)


def xs_pos(data, direction=1, quantile=50):
    # data = data[~np.isnan(data['FACTORVALUE'])].reset_index(drop=True)
    quantile_p = data['FACTORVALUE'].quantile(quantile / 100)
    quantile_n = data['FACTORVALUE'].quantile(1 - quantile / 100)
    data['POS'] = data['FACTORVALUE'].apply(
        lambda x: direction if x > quantile_p else (-direction if x < quantile_n else np.nan)
    )
    data = data[~np.isnan(data['POS'])].reset_index(drop=True)
    return xs_weight(data)


def tsmom(start_date, direction=1, **kwargs):
    data = kwargs['data']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']

    if direction == 1:
        index_name = 'tsmom_d' + str(window_days)
    else:
        index_name = 'tsrev_d' + str(window_days)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()
    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        lookback_date = calendar[i - window_days]
        data_t = data[
            np.array(data['TDATE'] == t_date) & np.array(data['VOL'] > kwargs['min_volume'])
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE']]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_lookback = data[data['TDATE'] == lookback_date][['TDATE', 'UCODE', 'CLOSE']]

        data_t = pd.merge(
            data_t[['EXCHANGE', 'PCODE', 'UCODE', 'CLOSE']],
            data_lookback[['UCODE', 'CLOSE']].rename(columns={'CLOSE': 'CLOSE0'}),
            on='UCODE', how='left'
        )

        data_t['FACTORVALUE'] = data_t['CLOSE'] / data_t['CLOSE0'] - 1

        data_tomorrow = data[np.array(data['TDATE'] == calendar[i + 1])]

        data_t = pd.merge(
            data_t, data_tomorrow[['UCODE', 'CLOSE']].rename(columns={'CLOSE': 'CLOSE1'}),
            on='UCODE', how='left'
        )

        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1)

        data_t['TDATE'] = calendar[i + 1]
        data_t['FACTOR'] = index_name
        data_t = ts_pos(data_t, direction=direction)

        positions = pd.concat([positions, data_t])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return
    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions.reset_index(drop=True)


def tsrev(start_date, **kwargs):
    return tsmom(start_date=start_date, direction=-1, **kwargs)


def tswr(start_date, price_field='CLOSE', lag=1, **kwargs):
    # 根据前一日仓单数据计算
    # 当日收盘价开平仓
    data = kwargs['data']
    data_wr = kwargs['data_wr']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    data_wr['UCODE'] = data_wr['EXCHANGE'].apply(lambda x: str(int(x))) + data_wr['PCODE'].apply(lambda x: str(int(x)))

    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']
    index_name = 'tswr_d' + str(window_days)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()
    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        data_wr_0 = data_wr[data_wr['TDATE'] == calendar[i - lag - window_days]][['UCODE', 'PNAME', 'WRQCURRENT']]
        data_wr_t = data_wr[data_wr['TDATE'] == calendar[i - lag]][['UCODE', 'WRQCURRENT']]

        data_wr_range = pd.merge(
            data_wr_0.rename(columns={'WRQCURRENT': 'WR0'}),
            data_wr_t.rename(columns={'WRQCURRENT': 'WR1'}),
            on='UCODE', how='left'
        )
        data_wr_range['FACTORVALUE'] = data_wr_range['WR1'] - data_wr_range['WR0']

        data_t = data[
            np.array(data['TDATE'] == t_date)
            & np.array(data['VOL'] > kwargs['min_volume'])
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', price_field]]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
        ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_tomorrow = data[
            np.array(data['TDATE'] == calendar[i + 1])
        ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', price_field]]

        data_t = pd.merge(data_t[['UCODE', price_field]], data_wr_range, on='UCODE', how='inner')

        data_t = pd.merge(
            data_t.rename(columns={price_field: 'CLOSE'}),
            data_tomorrow.rename(columns={price_field: 'CLOSE1'}),
            on='UCODE', how='left'
        )
        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1)
        data_t['TDATE'] = calendar[i + 1]
        data_t['FACTOR'] = index_name

        data_t = ts_pos(data=data_t, direction=-1)
        positions = pd.concat([positions, data_t])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return
    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    positions = positions[
        ['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'CLOSE1', 'POS', 'RETURN', 'FACTORVALUE', 'FACTOR', 'WEIGHT']
    ].reset_index(drop=True)
    return positions


def carry(start_date, **kwargs):
    data = kwargs['data']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']
    quantile = kwargs['quantile']
    index_name = 'carry_d' + str(window_days) + '_q' + str(quantile)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()
    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        lookback_date = calendar[i - window_days]
        data_ry = data[
            np.array(data['TDATE'] <= t_date)
            & np.array(data['TDATE'] > lookback_date)
            ].groupby('UCODE').mean().reset_index()[['UCODE', 'RY']]
        data_t = data[
            np.array(data['TDATE'] == t_date)
            & np.array(data['VOL'] > kwargs['min_volume'])
            ]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_t = pd.merge(data_t.rename(columns={'RY': 'RY0'}), data_ry, on='UCODE')
        data_t = data_t.rename(columns={'RY': 'FACTORVALUE'})
        data_tomorrow = data[
            np.array(data['TDATE'] == calendar[i + 1])
        ]

        # quantile_p = data_t['RY'].quantile(quantile / 100)
        # quantile_n = data_t['RY'].quantile(1 - quantile / 100)
        # data_t['POS'] = data_t['RY'].apply(lambda x: 1 if x > quantile_p else (-1 if x < quantile_n else 0))
        data_t = pd.merge(
            data_t, data_tomorrow[['UCODE', 'CLOSE']].rename(columns={'CLOSE': 'CLOSE1'}), on='UCODE', how='left'
        )
        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1)
        data_t = xs_pos(data=data_t, quantile=quantile)
        pos = data_t

        pos['TDATE'] = calendar[i + 1]
        pos['FACTOR'] = index_name

        positions = pd.concat([positions, pos])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return

    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions[
        ['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'CLOSE1', 'POS', 'RETURN', 'FACTORVALUE', 'FACTOR', 'WEIGHT']
    ].reset_index(drop=True)


def tscarry(start_date, **kwargs):
    data = kwargs['data']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']
    index_name = 'tscarry_d' + str(window_days)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()
    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        lookback_date = calendar[i - window_days]
        data_ry = data[
            np.array(data['TDATE'] <= t_date)
            & np.array(data['TDATE'] > lookback_date)
            ].groupby('UCODE').mean().reset_index()[['UCODE', 'RY']]
        data_t = data[
            np.array(data['TDATE'] == t_date)
            & np.array(data['VOL'] > kwargs['min_volume'])
            ]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_t = pd.merge(data_t.rename(columns={'RY': 'RY0'}), data_ry, on='UCODE')
        data_t = data_t.rename(columns={'RY': 'FACTORVALUE'})
        data_tomorrow = data[
            np.array(data['TDATE'] == calendar[i + 1])
        ]

        data_t = pd.merge(
            data_t, data_tomorrow[['UCODE', 'CLOSE']].rename(columns={'CLOSE': 'CLOSE1'}), on='UCODE', how='left'
        )
        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1)
        data_t = ts_pos(data=data_t)
        pos = data_t

        pos['TDATE'] = calendar[i + 1]
        pos['FACTOR'] = index_name

        positions = pd.concat([positions, pos])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return

    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions[
        ['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'CLOSE1', 'POS', 'RETURN', 'FACTORVALUE', 'FACTOR', 'WEIGHT']
    ].reset_index(drop=True)


def tsbasismom(start_date, **kwargs):
    data = kwargs['data']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']
    index_name = 'tsbasismom_d' + str(window_days)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()
    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        lookback_date = calendar[i - window_days]

        data_t = data[
            np.array(data['TDATE'] == t_date)
            & np.array(data['VOL'] > kwargs['min_volume'])
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CCODE', 'CCODE2', ]]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_t_1 = kwargs['data_raw'][kwargs['data_raw']['TDATE'] == lookback_date]

        data_t = data_t.merge(
            kwargs['data_raw'][['TDATE', 'CCODE', 'CLOSE', 'LDATE']].rename(
                columns={
                    # 'CLOSE': 'CLOSE1t0',
                    'LDATE': 'LDATE1',
                }
            ),
            on=['TDATE', 'CCODE'],
            how='left'
        ).merge(
            kwargs['data_raw'][['TDATE', 'CCODE', 'CLOSE', 'LDATE']].rename(
                columns={
                    'CLOSE': 'CLOSE2t0',
                    'CCODE': 'CCODE2',
                    'LDATE': 'LDATE2',
                }
            ),
            on=['TDATE', 'CCODE2'],
            how='left'
        ).merge(
            data_t_1[['CCODE', 'CLOSE']].rename(
                columns={
                    'CLOSE': 'CLOSE1t1'
                }
            ),
            on=['CCODE'],
            how='left'
        ).merge(
            data_t_1[['CCODE', 'CLOSE']].rename(
                columns={
                    'CLOSE': 'CLOSE2t1',
                    'CCODE': 'CCODE2'
                }
            ),
            on=['CCODE2'],
            how='left'
        )

        data_t = data_t.drop_duplicates(subset=['UCODE'], keep='last')
        data_t['return1'] = (data_t['CLOSE'] - data_t['CLOSE1t1']) / data_t['CLOSE1t1']
        data_t['return2'] = (data_t['CLOSE2t0'] - data_t['CLOSE2t1']) / data_t['CLOSE2t1']
        data_t['FACTORVALUE'] = (
                                        data_t['return1'] - data_t['return2']
                                ) * (data_t['LDATE2'] - data_t['LDATE1']).apply(lambda x: x.days)

        data_tomorrow = kwargs['data_raw'][
            np.array(kwargs['data_raw']['TDATE'] == calendar[i + 1])
        ]
        data_tomorrow = data_tomorrow.drop_duplicates(subset=['CCODE'], keep='last')
        data_t = pd.merge(
            data_t, data_tomorrow[['CCODE', 'CLOSE']].rename(columns={'CLOSE': 'CLOSE1'}), on='CCODE', how='left'
        )
        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1)
        data_t = ts_pos(data_t)
        pos = data_t[
            ['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'POS', 'FACTORVALUE', 'CLOSE1', 'RETURN', 'WEIGHT']
        ].copy()

        pos['TDATE'] = calendar[i + 1]
        pos['FACTOR'] = index_name

        positions = pd.concat([positions, pos])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return

    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions[
        ['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'CLOSE1', 'POS', 'RETURN', 'FACTORVALUE', 'FACTOR', 'WEIGHT']
    ].reset_index(drop=True)


def xsbasismom(start_date, **kwargs):
    data = kwargs['data']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']
    quantile = kwargs['quantile']
    index_name = 'xsbasismom_d' + str(window_days) + '_q' + str(quantile)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()
    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        lookback_date = calendar[i - window_days]

        data_t = data[
            np.array(data['TDATE'] == t_date)
            & np.array(data['VOL'] > kwargs['min_volume'])
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CCODE', 'CCODE2', ]]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_t_1 = kwargs['data_raw'][kwargs['data_raw']['TDATE'] == lookback_date]

        data_t = data_t.merge(
            kwargs['data_raw'][['TDATE', 'CCODE', 'CLOSE', 'LDATE']].rename(
                columns={
                    # 'CLOSE': 'CLOSE1t0',
                    'LDATE': 'LDATE1',
                }
            ),
            on=['TDATE', 'CCODE'],
            how='left'
        ).merge(
            kwargs['data_raw'][['TDATE', 'CCODE', 'CLOSE', 'LDATE']].rename(
                columns={
                    'CLOSE': 'CLOSE2t0',
                    'CCODE': 'CCODE2',
                    'LDATE': 'LDATE2',
                }
            ),
            on=['TDATE', 'CCODE2'],
            how='left'
        ).merge(
            data_t_1[['CCODE', 'CLOSE']].rename(
                columns={
                    'CLOSE': 'CLOSE1t1'
                }
            ),
            on=['CCODE'],
            how='left'
        ).merge(
            data_t_1[['CCODE', 'CLOSE']].rename(
                columns={
                    'CLOSE': 'CLOSE2t1',
                    'CCODE': 'CCODE2'
                }
            ),
            on=['CCODE2'],
            how='left'
        )

        data_t = data_t.drop_duplicates(subset=['UCODE'], keep='last')
        data_t['return1'] = (data_t['CLOSE'] - data_t['CLOSE1t1']) / data_t['CLOSE1t1']
        data_t['return2'] = (data_t['CLOSE2t0'] - data_t['CLOSE2t1']) / data_t['CLOSE2t1']
        data_t['FACTORVALUE'] = (
                                        data_t['return1'] - data_t['return2']
                                ) * (data_t['LDATE2'] - data_t['LDATE1']).apply(lambda x: x.days)
        data_t = data_t[~np.isinf(data_t['FACTORVALUE'])].reset_index(drop=True)

        # data_t['POS'] = data_t['FACTORVALUE'] / abs(data_t['FACTORVALUE'])
        # data_t = data_t[~np.isnan(data_t['POS'])].reset_index(drop=True)

        data_tomorrow = kwargs['data_raw'][
            np.array(kwargs['data_raw']['TDATE'] == calendar[i + 1])
        ]
        data_tomorrow = data_tomorrow.drop_duplicates(subset=['CCODE'], keep='last')
        data_t = pd.merge(
            data_t, data_tomorrow[['CCODE', 'CLOSE']].rename(columns={'CLOSE': 'CLOSE1'}), on='CCODE', how='left'
        )
        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1)
        data_t = xs_pos(data_t, quantile=quantile)
        pos = data_t[
            ['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'POS', 'FACTORVALUE', 'CLOSE1', 'RETURN', 'WEIGHT']
        ].copy()

        pos['TDATE'] = calendar[i + 1]
        pos['FACTOR'] = index_name

        positions = pd.concat([positions, pos])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return

    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions[
        ['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'CLOSE1', 'POS', 'RETURN', 'FACTORVALUE', 'FACTOR', 'WEIGHT']
    ].reset_index(drop=True)


def xsmom(start_date, direction=1, **kwargs):
    data = kwargs['data']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']
    quantile = kwargs['quantile']

    if direction == 1:
        index_name = 'xsmom_d' + str(window_days) + '_q' + str(quantile)
    else:
        index_name = 'xsrev_d' + str(window_days) + '_q' + str(quantile)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()

    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        lookback_date = calendar[i - window_days]
        data_t = data[
            np.array(data['TDATE'] == t_date) & np.array(data['VOL'] > kwargs['min_volume'])
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE']]
        data_lookback = data[data['TDATE'] == lookback_date][['TDATE', 'UCODE', 'CLOSE']]
        data_tomorrow = data[
            np.array(data['TDATE'] == calendar[i + 1])
        ]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_t = pd.merge(
            data_t[['EXCHANGE', 'PCODE', 'UCODE', 'CLOSE']],
            data_lookback[['UCODE', 'CLOSE']].rename(columns={'CLOSE': 'CLOSE0'}),
            on='UCODE', how='left'
        )
        data_t['FACTORVALUE'] = data_t['CLOSE'] / data_t['CLOSE0'] - 1

        data_t = pd.merge(
            data_t, data_tomorrow[['UCODE', 'CLOSE']].rename(columns={'CLOSE': 'CLOSE1'}), on='UCODE', how='left'
        )
        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1)
        data_t = xs_pos(data=data_t, direction=direction, quantile=quantile)
        pos = data_t

        pos['TDATE'] = calendar[i + 1]
        pos['FACTOR'] = index_name
        positions = pd.concat([positions, pos])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return
    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions.reset_index(drop=True)


def xsrev(start_date, **kwargs):
    return xsmom(start_date=start_date, direction=-1, **kwargs)


def xswr(start_date, price_field='CLOSE', lag=1, **kwargs):
    # 根据前一日仓单数据计算
    # 当日收盘价开平仓
    data = kwargs['data']
    data_wr = kwargs['data_wr']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    data_wr['UCODE'] = data_wr['EXCHANGE'].apply(lambda x: str(int(x))) + data_wr['PCODE'].apply(lambda x: str(int(x)))

    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']
    quantile = kwargs['quantile']
    index_name = 'xswr_d' + str(window_days) + '_q' + str(quantile)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()
    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        data_wr_0 = data_wr[data_wr['TDATE'] == calendar[i - lag - window_days]][['UCODE', 'PNAME', 'WRQCURRENT']]
        data_wr_t = data_wr[data_wr['TDATE'] == calendar[i - lag]][['UCODE', 'WRQCURRENT']]

        data_wr_range = pd.merge(
            data_wr_0.rename(columns={'WRQCURRENT': 'WR0'}),
            data_wr_t.rename(columns={'WRQCURRENT': 'WR1'}),
            on='UCODE', how='left'
        )
        data_wr_range['FACTORVALUE'] = (
            data_wr_range[data_wr_range['WR0'] > 0]['WR1'] / data_wr_range[data_wr_range['WR0'] > 0]['WR0'] - 1
        )

        data_t = data[
            np.array(data['TDATE'] == t_date) & np.array(data['VOL'] > kwargs['min_volume'])
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', price_field]]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_tomorrow = data[
            np.array(data['TDATE'] == calendar[i + 1])
        ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', price_field]]

        data_t = pd.merge(data_t[['UCODE', price_field]], data_wr_range, on='UCODE', how='inner')

        data_t = pd.merge(
            data_t.rename(columns={price_field: 'CLOSE'}),
            data_tomorrow.rename(columns={price_field: 'CLOSE1'}),
            on='UCODE', how='left'
        )

        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1)

        data_t['TDATE'] = calendar[i + 1]
        data_t['FACTOR'] = index_name
        data_t = xs_pos(data=data_t, direction=-1, quantile=quantile)
        positions = pd.concat([positions, data_t])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return
    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    positions = positions[
        ['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'CLOSE1', 'POS', 'RETURN', 'FACTORVALUE', 'FACTOR', 'WEIGHT']
    ].reset_index(drop=True)
    return positions


# n日净持仓
# 净头寸区间总和为正做多
def mr(start_date, price_field='CLOSE', lag=1, **kwargs):
    data = kwargs['data']
    data_mr = kwargs['data_mr']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    data_mr['UCODE'] = data_mr['EXCHANGE'].apply(lambda x: str(x)) + data_mr['PCODE'].apply(lambda x: str(x))

    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']
    index_name = 'mr_d' + str(window_days)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()
    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]

        data_mr_t = data_mr[
            np.array(data_mr['TDATE'] >= calendar[i - lag - window_days])
            & np.array(data_mr['TDATE'] <= calendar[i - lag])
            ].groupby(by=['UCODE']).sum()[['MR']].reset_index()
        data_mr_range = data_mr_t

        data_mr_range['FACTORVALUE'] = data_mr_range['MR']
        # data_mr_range['POS'] = data_mr_range['FACTORVALUE'] / data_mr_range['FACTORVALUE'].abs()
        # data_mr_range['POS'] = data_mr_range['POS'].fillna(0)
        data_t = data[
            np.array(data['TDATE'] == t_date) & np.array(data['VOL'] > kwargs['min_volume'])
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', price_field]]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_tomorrow = data[
            np.array(data['TDATE'] == calendar[i + 1])
        ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', price_field]]

        data_t = pd.merge(data_t[['UCODE', price_field]], data_mr_range, on='UCODE', how='inner')

        data_t = pd.merge(
            data_t.rename(columns={price_field: 'CLOSE'}),
            data_tomorrow.rename(columns={price_field: 'CLOSE1'}),
            on='UCODE', how='left'
        )

        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1) #* data_t['POS']

        # index_all.append(index_all[-1] * (1 + data_t['RETURN'].mean()))
        data_t['TDATE'] = calendar[i + 1]
        data_t['FACTOR'] = index_name
        data_t = ts_pos(data=data_t)
        positions = pd.concat([positions, data_t])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return
    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    # positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions[
        ['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'CLOSE1', 'POS', 'RETURN', 'FACTORVALUE', 'FACTOR', 'WEIGHT']
    ].reset_index(drop=True)


# n日净持仓
# 净头寸区间总和排序
def xsmr(start_date, price_field='CLOSE', lag=1, **kwargs):
    data = kwargs['data']
    data_mr = kwargs['data_mr']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    data_mr['UCODE'] = data_mr['EXCHANGE'].apply(lambda x: str(x)) + data_mr['PCODE'].apply(lambda x: str(x))

    calendar = data['TDATE'].drop_duplicates().tolist()

    quantile = kwargs['quantile']
    window_days = kwargs['window_days']
    index_name = 'xsmr_d' + str(window_days) + '_q' + str(quantile)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()
    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]

        data_mr_t = data_mr[
            np.array(data_mr['TDATE'] >= calendar[i - lag - window_days])
            & np.array(data_mr['TDATE'] <= calendar[i - lag])
            ].groupby(by=['UCODE']).sum()[['MR']].reset_index()
        data_mr_range = data_mr_t

        data_mr_range['FACTORVALUE'] = data_mr_range['MR']

        data_t = data[
            np.array(data['TDATE'] == t_date) & np.array(data['VOL'] > kwargs['min_volume'])
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', price_field]]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_tomorrow = data[
            np.array(data['TDATE'] == calendar[i + 1])
        ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', price_field]]

        data_t = pd.merge(data_t[['UCODE', price_field]], data_mr_range, on='UCODE', how='inner')

        data_t = pd.merge(
            data_t.rename(columns={price_field: 'CLOSE'}),
            data_tomorrow.rename(columns={price_field: 'CLOSE1'}),
            on='UCODE', how='left'
        )

        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1) #* data_t['POS']

        # index_all.append(index_all[-1] * (1 + data_t['RETURN'].mean()))
        data_t['TDATE'] = calendar[i + 1]
        data_t['FACTOR'] = index_name
        data_t = xs_pos(data=data_t, quantile=quantile)
        positions = pd.concat([positions, data_t])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return
    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    # positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions[
        ['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'CLOSE1', 'POS', 'RETURN', 'FACTORVALUE', 'FACTOR', 'WEIGHT']
    ].reset_index(drop=True)


# n日净持仓变化
# 净头寸时序增加做多
def tsmr(start_date, price_field='CLOSE', lag=1, **kwargs):
    data = kwargs['data']
    data_mr = kwargs['data_mr']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    data_mr['UCODE'] = data_mr['EXCHANGE'].apply(lambda x: str(x)) + data_mr['PCODE'].apply(lambda x: str(x))

    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']
    index_name = 'tsmr_d' + str(window_days)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()
    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        data_mr_0 = data_mr[
            data_mr['TDATE'] == calendar[i - lag - window_days]
        ].groupby(by=['UCODE']).sum()[['MR']].reset_index()
        data_mr_t = data_mr[
            data_mr['TDATE'] == calendar[i - lag]
        ].groupby(by=['UCODE']).sum()[['MR']].reset_index()

        data_mr_range = pd.merge(
            data_mr_0.rename(columns={'MR': 'MR0'}),
            data_mr_t.rename(columns={'MR': 'MR1'}),
            on='UCODE', how='left'
        )
        data_mr_range['FACTORVALUE'] = data_mr_range['MR1'] - data_mr_range['MR0']
        # data_mr_range['POS'] = data_mr_range['FACTORVALUE'] / data_mr_range['FACTORVALUE'].abs()
        # data_mr_range['POS'] = data_mr_range['POS'].fillna(0)

        data_t = data[
            np.array(data['TDATE'] == t_date) & np.array(data['VOL'] > kwargs['min_volume'])
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', price_field]]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_tomorrow = data[
            np.array(data['TDATE'] == calendar[i + 1])
        ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', price_field]]

        data_t = pd.merge(data_t[['UCODE', price_field]], data_mr_range, on='UCODE', how='inner')

        data_t = pd.merge(
            data_t.rename(columns={price_field: 'CLOSE'}),
            data_tomorrow.rename(columns={price_field: 'CLOSE1'}),
            on='UCODE', how='left'
        )

        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1) #* data_t['POS']

        # index_all.append(index_all[-1] * (1 + data_t['RETURN'].mean()))
        data_t['TDATE'] = calendar[i + 1]
        data_t['FACTOR'] = index_name
        data_t = ts_pos(data=data_t)
        positions = pd.concat([positions, data_t])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return
    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    # positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions[
        ['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'CLOSE1', 'POS', 'RETURN', 'FACTORVALUE', 'FACTOR', 'WEIGHT']
    ].reset_index(drop=True)


# n日净持仓变化
# 净头寸时序增加率排序
def xstsmr(start_date, price_field='CLOSE', lag=1, **kwargs):
    data = kwargs['data']
    data_mr = kwargs['data_mr']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    data_mr['UCODE'] = data_mr['EXCHANGE'].apply(lambda x: str(x)) + data_mr['PCODE'].apply(lambda x: str(x))

    calendar = data['TDATE'].drop_duplicates().tolist()

    quantile = kwargs['quantile']
    window_days = kwargs['window_days']
    index_name = 'xstsmr_d' + str(window_days) + '_q' + str(quantile)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()
    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        data_mr_0 = data_mr[
            data_mr['TDATE'] == calendar[i - lag - window_days]
            ].groupby(by=['UCODE']).sum()[['MR']].reset_index()
        data_mr_t = data_mr[
            data_mr['TDATE'] == calendar[i - lag]
            ].groupby(by=['UCODE']).sum()[['MR']].reset_index()

        data_mr_range = pd.merge(
            data_mr_0.rename(columns={'MR': 'MR0'}),
            data_mr_t.rename(columns={'MR': 'MR1'}),
            on='UCODE', how='left'
        )
        # data_mr_range['FACTORVALUE'] = data_mr_range['MR1'] - data_mr_range['MR0']
        data_mr_range['FACTORVALUE'] = (
                data_mr_range[data_mr_range['MR0'] > 0]['MR1'] / data_mr_range[data_mr_range['MR0'] > 0]['MR0'] - 1
        )

        data_t = data[
            np.array(data['TDATE'] == t_date) & np.array(data['VOL'] > kwargs['min_volume'])
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', price_field]]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_tomorrow = data[
            np.array(data['TDATE'] == calendar[i + 1])
        ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', price_field]]

        data_t = pd.merge(data_t[['UCODE', price_field]], data_mr_range, on='UCODE', how='inner')

        data_t = pd.merge(
            data_t.rename(columns={price_field: 'CLOSE'}),
            data_tomorrow.rename(columns={price_field: 'CLOSE1'}),
            on='UCODE', how='left'
        )

        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1) #* data_t['POS']

        # index_all.append(index_all[-1] * (1 + data_t['RETURN'].mean()))
        data_t['TDATE'] = calendar[i + 1]
        data_t['FACTOR'] = index_name
        data_t = xs_pos(data=data_t, quantile=quantile)
        positions = pd.concat([positions, data_t])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return
    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    # positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions[
        ['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'CLOSE1', 'POS', 'RETURN', 'FACTORVALUE', 'FACTOR', 'WEIGHT']
    ].reset_index(drop=True)


# n日净增减
# 净变化为正做多
def mrchg(start_date, price_field='CLOSE', lag=1, **kwargs):
    data = kwargs['data']
    data_mr = kwargs['data_mr']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    data_mr['UCODE'] = data_mr['EXCHANGE'].apply(lambda x: str(x)) + data_mr['PCODE'].apply(lambda x: str(x))

    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']
    index_name = 'mrchg_d' + str(window_days)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()
    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        data_mr_t = data_mr[
            np.array(data_mr['TDATE'] >= calendar[i - lag - window_days])
            & np.array(data_mr['TDATE'] <= calendar[i - lag])
            ].groupby(by=['UCODE']).sum()[['MRCHG']].reset_index()
        data_mr_range = data_mr_t

        data_mr_range['FACTORVALUE'] = data_mr_range['MRCHG']
        # data_mr_range['POS'] = data_mr_range['FACTORVALUE'] / data_mr_range['FACTORVALUE'].abs()
        # data_mr_range['POS'] = data_mr_range['POS'].fillna(0)

        data_t = data[
            np.array(data['TDATE'] == t_date) & np.array(data['VOL'] > kwargs['min_volume'])
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', price_field]]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_tomorrow = data[
            np.array(data['TDATE'] == calendar[i + 1])
        ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', price_field]]

        data_t = pd.merge(data_t[['UCODE', price_field]], data_mr_range, on='UCODE', how='inner')

        data_t = pd.merge(
            data_t.rename(columns={price_field: 'CLOSE'}),
            data_tomorrow.rename(columns={price_field: 'CLOSE1'}),
            on='UCODE', how='left'
        )

        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1) #* data_t['POS']

        # index_all.append(index_all[-1] * (1 + data_t['RETURN'].mean()))
        data_t['TDATE'] = calendar[i + 1]
        data_t['FACTOR'] = index_name
        data_t = ts_pos(data=data_t)
        positions = pd.concat([positions, data_t])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return
    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    # positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions[
        ['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'CLOSE1', 'POS', 'RETURN', 'FACTORVALUE', 'FACTOR', 'WEIGHT']
    ].reset_index(drop=True)


# n日净增减
# 净变化排序
def xsmrchg(start_date, price_field='CLOSE', lag=1, **kwargs):
    data = kwargs['data']
    data_mr = kwargs['data_mr']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    data_mr['UCODE'] = data_mr['EXCHANGE'].apply(lambda x: str(x)) + data_mr['PCODE'].apply(lambda x: str(x))

    calendar = data['TDATE'].drop_duplicates().tolist()

    quantile = kwargs['quantile']
    window_days = kwargs['window_days']
    index_name = 'xsmrchg_d' + str(window_days) + '_q' + str(quantile)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()
    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        data_mr_t = data_mr[
            np.array(data_mr['TDATE'] >= calendar[i - lag - window_days])
            & np.array(data_mr['TDATE'] <= calendar[i - lag])
            ].groupby(by=['UCODE']).sum()[['MRCHG']].reset_index()
        data_mr_range = data_mr_t

        data_mr_range['FACTORVALUE'] = data_mr_range['MRCHG']


        data_t = data[
            np.array(data['TDATE'] == t_date) & np.array(data['VOL'] > kwargs['min_volume'])
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', price_field]]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_tomorrow = data[
            np.array(data['TDATE'] == calendar[i + 1])
        ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', price_field]]

        data_t = pd.merge(data_t[['UCODE', price_field]], data_mr_range, on='UCODE', how='inner')

        data_t = pd.merge(
            data_t.rename(columns={price_field: 'CLOSE'}),
            data_tomorrow.rename(columns={price_field: 'CLOSE1'}),
            on='UCODE', how='left'
        )

        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1) #* data_t['POS']

        # index_all.append(index_all[-1] * (1 + data_t['RETURN'].mean()))
        data_t['TDATE'] = calendar[i + 1]
        data_t['FACTOR'] = index_name
        data_t = xs_pos(data=data_t, quantile=quantile)
        positions = pd.concat([positions, data_t])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return
    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    # positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions[
        ['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'CLOSE1', 'POS', 'RETURN', 'FACTORVALUE', 'FACTOR', 'WEIGHT']
    ].reset_index(drop=True)


def tspoichg(start_date, **kwargs):
    data = kwargs['data']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']

    index_name = 'tspoichg_d' + str(window_days)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()
    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        lookback_date = calendar[i - window_days]
        data_t = data[
            np.array(data['TDATE'] == t_date)
            & np.array(data['VOL'] > kwargs['min_volume'])
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'POI']]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_lookback = data[
            np.array(data['TDATE'] == lookback_date)
            & np.array(data['POI'] > 0)
        ][['TDATE', 'UCODE', 'CLOSE', 'POI']]

        data_t = pd.merge(
            data_t[['EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'POI']],
            data_lookback[['UCODE', 'POI']].rename(columns={'POI': 'POI0'}),
            on='UCODE', how='left'
        )

        data_t['FACTORVALUE'] = data_t['POI'] / data_t['POI0'] - 1

        data_tomorrow = data[np.array(data['TDATE'] == calendar[i + 1])]

        data_t = pd.merge(
            data_t, data_tomorrow[['UCODE', 'CLOSE']].rename(columns={'CLOSE': 'CLOSE1'}),
            on='UCODE', how='left'
        )

        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1)

        data_t['TDATE'] = calendar[i + 1]
        data_t['FACTOR'] = index_name

        data_t = ts_pos(data_t)

        positions = pd.concat([positions, data_t])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return
    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions.drop(columns=['POI', 'POI0']).reset_index(drop=True)


def tspvolchg(start_date, **kwargs):
    data = kwargs['data']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']

    index_name = 'tspvolchg_d' + str(window_days)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()
    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        lookback_date = calendar[i - window_days]
        data_t = data[
            np.array(data['TDATE'] == t_date)
            & np.array(data['VOL'] > kwargs['min_volume'])
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'PVOL']]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_lookback = data[
            np.array(data['TDATE'] == lookback_date)
            & np.array(data['PVOL'] > 0)
            ][['TDATE', 'UCODE', 'CLOSE', 'PVOL']]

        data_t = pd.merge(
            data_t[['EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'PVOL']],
            data_lookback[['UCODE', 'PVOL']].rename(columns={'PVOL': 'PVOL0'}),
            on='UCODE', how='left'
        )

        data_t['FACTORVALUE'] = data_t['PVOL'] / data_t['PVOL0'] - 1

        data_tomorrow = data[np.array(data['TDATE'] == calendar[i + 1])]

        data_t = pd.merge(
            data_t, data_tomorrow[['UCODE', 'CLOSE']].rename(columns={'CLOSE': 'CLOSE1'}),
            on='UCODE', how='left'
        )

        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1)

        data_t['TDATE'] = calendar[i + 1]
        data_t['FACTOR'] = index_name

        data_t = ts_pos(data_t)

        positions = pd.concat([positions, data_t])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return
    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions.drop(columns=['PVOL', 'PVOL0']).reset_index(drop=True)


def xspvolchg(start_date, **kwargs):
    data = kwargs['data']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']
    quantile = kwargs['quantile']

    index_name = 'xspvolchg_d' + str(window_days) + '_q' + str(quantile)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()
    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        lookback_date = calendar[i - window_days]
        data_t = data[
            np.array(data['TDATE'] == t_date)
            & np.array(data['VOL'] > kwargs['min_volume'])
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'PVOL']]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_lookback = data[
            np.array(data['TDATE'] == lookback_date)
            & np.array(data['PVOL'] > 0)
            ][['TDATE', 'UCODE', 'CLOSE', 'PVOL']]

        data_t = pd.merge(
            data_t[['EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'PVOL']],
            data_lookback[['UCODE', 'PVOL']].rename(columns={'PVOL': 'PVOL0'}),
            on='UCODE', how='left'
        )

        data_t['FACTORVALUE'] = data_t['PVOL'] / data_t['PVOL0'] - 1

        data_tomorrow = data[np.array(data['TDATE'] == calendar[i + 1])]

        data_t = pd.merge(
            data_t, data_tomorrow[['UCODE', 'CLOSE']].rename(columns={'CLOSE': 'CLOSE1'}),
            on='UCODE', how='left'
        )

        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1)

        data_t['TDATE'] = calendar[i + 1]
        data_t['FACTOR'] = index_name

        data_t = xs_pos(data_t, quantile=quantile)

        positions = pd.concat([positions, data_t])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return
    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions.drop(columns=['PVOL', 'PVOL0']).reset_index(drop=True)


def xspoichg(start_date, **kwargs):
    data = kwargs['data']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']
    quantile = kwargs['quantile']

    index_name = 'xspoichg_d' + str(window_days) + '_q' + str(quantile)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()

    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        lookback_date = calendar[i - window_days]
        data_t = data[
            np.array(data['TDATE'] == t_date)
            & np.array(data['VOL'] > kwargs['min_volume'])
        ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'POI']]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_lookback = data[
            np.array(data['TDATE'] == lookback_date)
            & np.array(data['POI'] > 0)
            ][['TDATE', 'UCODE', 'CLOSE', 'POI']]

        data_t = pd.merge(
            data_t[['EXCHANGE', 'PCODE', 'UCODE', 'CLOSE', 'POI']],
            data_lookback[['UCODE', 'POI']].rename(columns={'POI': 'POI0'}),
            on='UCODE', how='left'
        )

        data_t['FACTORVALUE'] = data_t['POI'] / data_t['POI0'] - 1

        data_tomorrow = data[np.array(data['TDATE'] == calendar[i + 1])]
        data_t = pd.merge(
            data_t, data_tomorrow[['UCODE', 'CLOSE']].rename(columns={'CLOSE': 'CLOSE1'}), on='UCODE', how='left'
        )
        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1)
        data_t = xs_pos(data=data_t, quantile=quantile)
        pos = data_t

        pos['TDATE'] = calendar[i + 1]
        pos['FACTOR'] = index_name
        positions = pd.concat([positions, pos])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return
    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions.drop(columns=['POI', 'POI0']).reset_index(drop=True)


def xssigma(start_date, **kwargs):
    data = kwargs['data']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']
    quantile = kwargs['quantile']

    index_name = 'xssigma_d' + str(window_days) + '_q' + str(quantile)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()
    data_close = data.pivot(index='TDATE', columns='UCODE', values='CLOSE')
    data_vol = data.pivot(index='TDATE', columns='UCODE', values='VOL')
    data_sigma = data_close.pct_change().rolling(window_days).std(ddof=1) * np.sqrt(240)

    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]

        ucode_vol = data_vol.columns[data_vol.loc[t_date, :] > kwargs['min_volume']].tolist()
        ucode_liq = data_vol.columns[
            data_vol.loc[calendar[i - max(liq_days, window_days)], :] > kwargs['min_volume']
        ].tolist()
        ucode = list(set(ucode_liq).intersection(set(ucode_vol)))
        data_t = data[
            np.array(data['TDATE'] == t_date)
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE']]

        data_t = data_t.merge(
            pd.DataFrame(data_sigma.loc[t_date, ucode]).rename(columns={t_date: 'FACTORVALUE'}).reset_index(),
            on='UCODE', how='inner'
        )

        data_tomorrow = data[np.array(data['TDATE'] == calendar[i + 1])]
        data_t = pd.merge(
            data_t, data_tomorrow[['UCODE', 'CLOSE']].rename(columns={'CLOSE': 'CLOSE1'}), on='UCODE', how='left'
        )
        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1)
        data_t = xs_pos(data=data_t, quantile=quantile)
        pos = data_t

        pos['TDATE'] = calendar[i + 1]
        pos['FACTOR'] = index_name
        positions = pd.concat([positions, pos])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return
    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions.reset_index(drop=True)


def xsskew(start_date, **kwargs):
    data = kwargs['data']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']
    quantile = kwargs['quantile']

    index_name = 'xsskew_d' + str(window_days) + '_q' + str(quantile)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()

    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        lookback_date = calendar[i - window_days]
        data_t = data[
            np.array(data['TDATE'] == t_date) & np.array(data['VOL'] > kwargs['min_volume'])
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE']]
        data_lookback = data[
            np.array(data['TDATE'] >= lookback_date) & np.array(data['TDATE'] <= t_date)
        ][['TDATE', 'UCODE', 'CLOSE']]
        data_lookback_skew = data_lookback.pivot(index='TDATE', columns='UCODE', values='CLOSE').skew().reset_index()
        data_tomorrow = data[
            np.array(data['TDATE'] == calendar[i + 1])
        ]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_t = pd.merge(
            data_t[['EXCHANGE', 'PCODE', 'UCODE', 'CLOSE']],
            data_lookback_skew[['UCODE', 0]].rename(columns={0: 'FACTORVALUE'}),
            on='UCODE', how='left'
        )
        # data_t['FACTORVALUE'] = data_t['CLOSE'] / data_t['CLOSE0'] - 1

        data_t = pd.merge(
            data_t, data_tomorrow[['UCODE', 'CLOSE']].rename(columns={'CLOSE': 'CLOSE1'}), on='UCODE', how='left'
        )
        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1)
        data_t = xs_pos(data=data_t, quantile=quantile)
        pos = data_t

        pos['TDATE'] = calendar[i + 1]
        pos['FACTOR'] = index_name
        positions = pd.concat([positions, pos])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return
    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions.reset_index(drop=True)


def tsskew(start_date, **kwargs):
    data = kwargs['data']
    liq_days = kwargs['liq_days']
    data['UCODE'] = data['EXCHANGE'].apply(lambda x: str(x)) + data['PCODE'].apply(lambda x: str(x))
    calendar = data['TDATE'].drop_duplicates().tolist()

    window_days = kwargs['window_days']

    index_name = 'tsskew_d' + str(window_days)

    start_index = calendar.index(np.array(calendar)[np.array(calendar) >= start_date][0])

    positions = pd.DataFrame()

    for i in range(start_index, len(calendar) - 1):
        t_date = calendar[i]
        lookback_date = calendar[i - window_days]
        data_t = data[
            np.array(data['TDATE'] == t_date) & np.array(data['VOL'] > kwargs['min_volume'])
            ][['TDATE', 'EXCHANGE', 'PCODE', 'UCODE', 'CLOSE']]
        data_lookback = data[
            np.array(data['TDATE'] >= lookback_date) & np.array(data['TDATE'] <= t_date)
            ][['TDATE', 'UCODE', 'CLOSE']]
        data_lookback_skew = data_lookback.pivot(index='TDATE', columns='UCODE', values='CLOSE').skew().reset_index()
        data_tomorrow = data[
            np.array(data['TDATE'] == calendar[i + 1])
        ]

        data_liq = data[
            data['TDATE'] == calendar[i - liq_days]
            ][['UCODE']]

        data_t = pd.merge(data_liq, data_t, on='UCODE', how='inner')

        data_t = pd.merge(
            data_t[['EXCHANGE', 'PCODE', 'UCODE', 'CLOSE']],
            data_lookback_skew[['UCODE', 0]].rename(columns={0: 'FACTORVALUE'}),
            on='UCODE', how='left'
        )
        # data_t['FACTORVALUE'] = data_t['CLOSE'] / data_t['CLOSE0'] - 1

        data_t = pd.merge(
            data_t, data_tomorrow[['UCODE', 'CLOSE']].rename(columns={'CLOSE': 'CLOSE1'}), on='UCODE', how='left'
        )
        data_t['RETURN'] = (data_t['CLOSE1'] / data_t['CLOSE'] - 1)
        data_t = ts_pos(data=data_t)
        pos = data_t

        pos['TDATE'] = calendar[i + 1]
        pos['FACTOR'] = index_name
        positions = pd.concat([positions, pos])
        print(index_name + ', ' + t_date.strftime('%Y-%m-%d') + ', ' + str(sum(data_t['POS'] != 0)))
    if len(positions) == 0:
        return
    positions['WEIGHT'] = positions['WEIGHT'] * 100
    positions['RETURN'] = positions['RETURN'] * 100
    positions['FACTORVALUE'] = positions['FACTORVALUE'] * 100
    return positions.reset_index(drop=True)



"""
线上和本地的Brinson归因结果校验
"""
import pandas as pd
import hbshare as hbs
from hbshare.fe.brinson_attr.equity_brinson_attribution import EquityBrinsonAttribution


class BrinsonAttrCheck:
    def __init__(self, fund_id, mode='single'):
        self.fund_id = fund_id
        self.mode = mode
        if self.mode == 'all':
            self._load_summary_info()
        self._load_data()

    def _load_summary_info(self):
        # fund_info
        sql_script = "select jjdm,clrq,zzrq, (select count(1) from st_fund.t_st_gm_rhb r where r.jjdm = a.jjdm ) " \
                     "hb_count from st_fund.t_st_gm_jjxx a where a.Jjfl = '{}' and m_opt_type<>'03' and " \
                     "clrq is not null order by clrq asc".format('1')
        res = hbs.db_data_query('funduser', sql_script, page_size=10000)
        equity_info = pd.DataFrame(res['data'])
        equity_info['category'] = 'equity'
        sql_script = "select jjdm,clrq,zzrq, (select count(1) from st_fund.t_st_gm_rhb r where r.jjdm = a.jjdm ) " \
                     "hb_count from st_fund.t_st_gm_jjxx a where a.Jjfl = '{}' and m_opt_type<>'03' and " \
                     "clrq is not null order by clrq asc".format('3')
        res = hbs.db_data_query('funduser', sql_script, page_size=10000)
        hybrid_info = pd.DataFrame(res['data'])
        hybrid_info['category'] = 'hybrid'

        fund_info = pd.concat([equity_info, hybrid_info]).rename(
            columns={"clrq": "start_date", "zzrq": "end_date", "jjdm": "fund_id"})

        fund_info['start_date'] = fund_info['start_date'].astype(str)
        fund_info['end_date'] = fund_info['end_date'].fillna('20210824').astype(str)
        print("--- All Equity and Hybrid Fund Num = {} ---".format(fund_info.shape[0]))

        # 剔除成立时间不足一年的
        fund_info = fund_info[fund_info['start_date'] < '20200701']
        print("--- After Operating Period Filter Num = {} ---".format(fund_info.shape[0]))

        self.fund_info = fund_info[['fund_id', 'category', 'start_date', 'end_date']]

    def _load_data(self):
        sql_script = "select jjdm,clrq,zzrq, (select count(1) from st_fund.t_st_gm_rhb r where r.jjdm = a.jjdm ) " \
                     "hb_count from st_fund.t_st_gm_jjxx a where a.Jjfl in ('1', '3') and m_opt_type<>'03' and " \
                     "clrq is not null and jjdm = '{}' order by clrq asc".format(self.fund_id)
        res = hbs.db_data_query('funduser', sql_script)
        df = pd.DataFrame(res['data']).set_index('jjdm')

        self.start_date = \
            str(df.loc[self.fund_id, 'clrq']) if str(df.loc[self.fund_id, 'clrq']) > '20100101' else '20091220'
        self.end_date = str(df.loc[self.fund_id, 'zzrq']) if df.loc[self.fund_id, 'zzrq'] is not None else '20210701'

        sql_script = "SELECT * FROM st_fund.r_st_hold_excess_attr_df where jjdm = '{}'".format(self.fund_id)
        res = hbs.db_data_query('funduser', sql_script)
        attr_df = pd.DataFrame(res['data']).rename(columns={"jjdm": "fund_id", "tjrq": "trade_date"})
        if not attr_df.empty:
            cols = ['trade_date', 'portfolio_return', 'benchmark_return', 'asset_allo', 'sector_allo',
                    'equity_selection', 'trading', 'portfolio_weight', 'benchmark_weight', 'fund_id']
            attr_df_ol = attr_df[cols].sort_values(by='trade_date')
        else:
            attr_df_ol = pd.DataFrame()

        sql_script = "SELECT * FROM st_fund.r_st_hold_sector_attr_df where jjdm = '{}'".format(self.fund_id)
        res = hbs.db_data_query('funduser', sql_script)
        sector_attr_df = pd.DataFrame(res['data']).rename(columns={"jjdm": "fund_id", "tjrq": "trade_date"})
        if not sector_attr_df.empty:
            cols = ['trade_date', 'industry', 'portfolio_weight', 'benchmark_weight', 'active_weight',
                    'portfolio_return', 'benchmark_return', 'active_return',
                    'sector_allo', 'equity_selection', 'management', 'fund_id']
            sector_attr_df_ol = sector_attr_df[cols].sort_values(by='trade_date')
        else:
            sector_attr_df_ol = pd.DataFrame()

        self.attr_result = {"realized_period_attr_df": attr_df_ol, "sector_attr_df": sector_attr_df_ol}

    def get_construct_result(self):
        # 本地结果
        attr_result = \
            EquityBrinsonAttribution(self.fund_id, '000906', self.start_date, self.end_date).get_attribution_result()

        return self.attr_result, attr_result


if __name__ == '__main__':
    res_ol, res_lc = BrinsonAttrCheck(fund_id='000031').get_construct_result()
    print(res_ol['realized_period_attr_df'])
    print("\n")
    print(res_lc['realized_period_attr_df'].round(6))
    # f_info = BrinsonAttrCheck(fund_id='110011', mode='all').fund_info
    # f_info = f_info[f_info['end_date'] > '20100101'].set_index('fund_id')
    #
    # from tqdm import tqdm
    #
    # for fund_id in tqdm(f_info.index.tolist()):
    #     start_date = f_info.loc[fund_id, 'start_date']
    #     start_date = start_date if start_date > '2010101' else '20091220'
    #     end_date = f_info.loc[fund_id, 'end_date']
    #     sql = "SELECT JJDM, JSRQ, ZQDM, ZJBL FROM funddb.GPZH WHERE JJDM = '{}' and GGRQ >= {} " \
    #           "and GGRQ <= {}".format(fund_id, start_date, end_date)
    #     res_outer = hbs.db_data_query('readonly', sql, page_size=10000)
    #     holding_df = pd.DataFrame(res_outer['data'])
    #     date_list = [x for x in holding_df['JSRQ'].unique() if x[4:6] in ["06", "12"]]
    #     f_info.loc[fund_id, 'holding_num'] = len(date_list)
    #
    # sql_script = "SELECT jjdm, count(*) as num FROM st_fund.r_st_hold_excess_attr_df group by jjdm"
    # res = hbs.db_data_query('funduser', sql_script, page_size=10000)
    # df = pd.DataFrame(res['data'])



#!/usr/bin/python
#coding:utf-8
import pandas as pd
import numpy as np
from sqlalchemy import create_engine
from hbshare.quant.Kevin.rm_associated import engine_params
from datetime import datetime
import pywt


start_date = '20120101'
end_date = '20200408'


sql_script = "SELECT * FROM turnover_structure WHERE TRADE_DATE >= {} and TRADE_DATE <= {}".format(start_date, end_date)
engine = create_engine(engine_params)
data = pd.read_sql(sql_script, engine)
data['trade_date'] = data['trade_date'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))

data = data[['trade_date', 'cr5', 'cr10', 'avg_mkt', 'close']]
db4 = pywt.Wavelet('db4')
coeffs = pywt.wavedec(np.array(data['cr5']), db4)
# coeffs[len(coeffs) - 1] *= 0
# coeffs[len(coeffs) - 2] *= 0
# coeffs[len(coeffs) - 3] *= 0
# coeffs[len(coeffs) - 4] *= 0
# coeffs[len(coeffs) - 5] *= 0
# coeffs[len(coeffs) - 6] *= 0

threshold = 0.9
for i in range(1, len(coeffs)):
    coeffs[i] = pywt.threshold(coeffs[i], threshold*max(coeffs[i]))  # 将噪声滤波

meta = pywt.waverec(coeffs, db4)
data['cr5_denoise'] = meta[:-1]
data.set_index('trade_date')[['cr5', 'cr5_denoise']].plot.line()


from sqlalchemy import create_engine
import pymysql


def create_table(table_name, sql_script):
    db = pymysql.connect(host='192.168.223.152', user='admin', password='mysql', database='riskmodel')
    cursor = db.cursor()
    cursor.execute("DROP TABLE IF EXISTS {}".format(table_name))

    try:
        cursor.execute(sql_script)
        db.commit()
        print("创建数据库成功!")
    except Exception as e:
        print("创建数据库失败: case {}".format(e))

    finally:
        cursor.close()
        db.close()


def delete_duplicate_records(sql_script):
    db = pymysql.connect(host='192.168.223.152', user='admin', password='mysql', database='riskmodel')
    cursor = db.cursor()

    try:
        cursor.execute(sql_script)
        db.commit()
        print("删除重复数据成功！")
    except Exception as e:
        print("删除重复数据失败: case {}".format(e))

    finally:
        cursor.close()
        db.close()


class WriteToDB:
    def __init__(self):
        self.engine = create_engine("mysql+pymysql://{}:{}@{}:{}/{}".format(
            'admin', 'mysql', '192.168.223.152', '3306', 'riskmodel'))

    def write_to_db(self, df, table_name):
        df.to_sql(table_name, self.engine, index=False, if_exists='append')



# -*- coding: utf-8 -*-

from collections import OrderedDict
import configparser
import io
import os
import pickle
import sys


class FileReader:
    @staticmethod
    def read_file_list(filename, encoding='utf-8'):
        to_ret = []
        if sys.version[0] == '3':
            with open(filename, encoding=encoding) as f:
                for line in f:
                    to_ret.append(line.strip())
        else:
            with io.open(filename, encoding=encoding) as f:
                for line in f:
                    to_ret.append(line.strip())
        return to_ret

    @staticmethod
    def read_file_property(filename, section_name, key):
        config = configparser.ConfigParser()
        config.read_file(filename, encoding='utf-8')
        if section_name not in config or key not in config[section_name]:
            return None
        return config[section_name][key]

    @staticmethod
    def get_json_cfg(filename):
        dir_path = os.path.dirname(__file__)
        cfg_path = os.path.join(dir_path, '../../cfg/' + filename)
        cfg_string = FileReader.get_file_content(cfg_path)
        return cfg_string

    @staticmethod
    def get_file_content(filename):
        file_object = open(filename, 'r', encoding='utf-8') if sys.version[0] == '3' else io.open(filename, 'r', encoding='utf-8')
        try:
            content = file_object.read()
        finally:
            file_object.close()
        return content

    @staticmethod
    def load_progress(file_name, equ_init_date):
        """
        读取个股处理进度字典，
        :param file_name: @format "stock start_date" or "stock"
        :param equ_init_date:
        :return: {stock:start_date}
        """
        stock_progress_list = FileReader.read_file_list(file_name)
        stock_progress_dict = OrderedDict()
        for line in stock_progress_list:
            items = line.split(',')
            stock_progress_dict[items[0]] = equ_init_date if len(items) == 1 else items[1]
        return stock_progress_dict

    @staticmethod
    def load_object_from_file(file_name):
        obj = pickle.load(open(file_name, 'rb'))
        return obj


#!/usr/bin/python
#coding:utf-8

"""
@author: Meng.lv
@contact: meng.lv@howbuy.com
@software: PyCharm
@file: data_query.py
@time: 2021/1/20 8:44
"""
import time
import hbshare as hbs
from hbshare.db.fund import cons as ct
import json


def db_data_save_raw(database, table_name,  data=[], private_key=[], retry_count=1, pause=0.01, timeout = 300):
    """
        通用DB数据保存接口
    :param database: 数据库名（不支持ORACLE库,支持(MYSQL)库: stfund(公募产品库),
            stportfolio(组合产品库),sthedge(私募产品库), stfixed(固收产品库),
            stmarket(市场指数库),stashare(市场股票库), stinsurance(保险产品库), stmain(公共库)
    :param table_name: 执行脚本
    :param data: 数据
    :param private_key: 业务主键
    :param retry_count: 重试次数
    :param pause: 重试间隔时间
    :param timeout: 超时时间 (0,600]
    :return:
    """
    api = hbs.hb_api(timeout=timeout)
    for _ in range(retry_count):
        time.sleep(pause)
        #ct._write_console()
        if hbs.is_prod_env():
            url = "http://fpc-data.intelnal.howbuy.com/data/modify/commonapi?dataTrack=%s" % (time.time())
        else:
            url = "http://192.168.223.48:8101/data/modify/commonapi?dataTrack=%s" % (time.time())
        post_body = {
                "database": database,
                "tableName": table_name,
                "privateKey": private_key
        }
        post_body["dataList"] = data
        data = api.query(url, 'post', post_body)
        # print(post_body)
        success = data['success']
        if not success:
            status = str(data['returnCode'])+":"+str(data['returnMsg'])
            raise ValueError(status)
        # data_frame = pd.DataFrame(data['data'])
        # pagination = {
        #     "pageNum": page_num,
        #     "pageSize": page_size,
        #     "total": int(data['total']),
        #     "pages": int(data['pages'])
        # }
        return success
    raise IOError(ct.NETWORK_URL_ERROR_MSG)

def db_data_save(database, table_name, dt, keys, limit=1000):
    """
            通用DB数据保存接口
        :param database: 数据库名（不支持ORACLE库,支持(MYSQL)库: stfund(公募产品库),
                stportfolio(组合产品库),sthedge(私募产品库), stfixed(固收产品库),
                stmarket(市场指数库),stashare(市场股票库), stinsurance(保险产品库), stmain(公共库)
        :param table_name: 执行脚本
        :param dt: 数据
        :param keys: 业务主键
        :return:
    """
    ret = True
    dt['M_OPT_TYPE'] = '01'
    dt.columns = dt.columns.str.upper()
    total = ( dt.shape[0] // limit  ) if dt.shape[0] % limit == 0 else dt.shape[0] // limit + 1
    for i in range(total):
        start = i * limit
        d = dt[ start : start + limit ]
        #print(d['commetrade'])
        json_data = json.loads(d.to_json(orient = 'records'))
        success = db_data_save_raw(database, table_name, json_data, keys)
        if not success:
            print("there are some error", d)
        ret = ret and success
    return ret


"""
私募指增产品2021年以来的净值归因
"""
import os
import pandas as pd
from datetime import datetime
import hbshare as hbs
from hbshare.fe.common.util.regressions import Regression


class PrivateFundNavAttr:
    def __init__(self, path, start_date, end_date, nav_frequency='day'):
        self.path = path
        self.start_date = start_date
        self.end_date = end_date
        self.nav_frequency = nav_frequency
        self._load_data()

    def _load_calendar(self):
        sql_script = "SELECT JYRQ, SFJJ, SFZM, SFYM FROM funddb.JYRL WHERE JYRQ >= {} and JYRQ <= {}".format(
            self.start_date, self.end_date)
        res = hbs.db_data_query('readonly', sql_script, page_size=5000)
        df = pd.DataFrame(res['data']).rename(
            columns={"JYRQ": 'calendarDate', "SFJJ": 'isOpen',
                     "SFZM": "isWeekEnd", "SFYM": "isMonthEnd"}).sort_values(by='calendarDate')
        df['isOpen'] = df['isOpen'].astype(int).replace({0: 1, 1: 0})
        df['isWeekEnd'] = df['isWeekEnd'].fillna(0).astype(int)
        df['isMonthEnd'] = df['isMonthEnd'].fillna(0).astype(int)

        trading_day_list = df[df['isWeekEnd'] == 1]['calendarDate'].tolist()

        return trading_day_list

    def _load_fund_nav_data(self):
        fund_info = pd.read_excel(os.path.join(self.path, '股票量化池.xlsx'), sheet_name=0, header=1)
        fund_info = fund_info.fillna(method='ffill')
        fund_info = fund_info[fund_info['策略类型'] == '指数增强'][['分类', '基金管理人', '基金简称', '基金代码', '净值状态']]
        fund_nav_dict = dict()
        # from sql
        fund_id_list = fund_info[fund_info['净值状态'] == 'DB']['基金代码'].tolist()
        for fund_id in fund_id_list:
            fund_name = fund_info.set_index('基金代码').loc[fund_id, '基金简称']
            sql_script = "SELECT a.jjdm fund_id, b.jzrq TRADEDATE, b.fqdwjz as ADJ_NAV from " \
                         "st_hedge.t_st_jjxx a, st_hedge.t_st_rhb b where a.cpfl = '4' and a.jjdm = b.jjdm " \
                         "and a.jjzt not in ('3') " \
                         "and a.jjdm = '{}' and b.jzrq >= {} and b.jzrq <= {} " \
                         "order by b.jzrq".format(fund_id, self.start_date, self.end_date)
            res = hbs.db_data_query("highuser", sql_script, page_size=5000)
            data = pd.DataFrame(res['data'])
            fund_nav_dict[fund_name] = data.set_index('TRADEDATE')['ADJ_NAV']
        # from excel
        raw_dict = pd.read_excel(os.path.join(self.path, '指增净值.xlsx'), sheet_name=None)
        for fund_name, nav_df in raw_dict.items():
            if type(nav_df['净值日期'].tolist()[0]) != str:
                nav_df['净值日期'] = nav_df['净值日期'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
            nav_df['净值日期'] = nav_df['净值日期'].apply(lambda x: x.replace('-', ''))
            nav_df = nav_df[
                (nav_df['净值日期'] >= self.start_date) & (nav_df['净值日期'] <= self.end_date)].sort_values(by='净值日期')
            fund_nav_dict[fund_name] = nav_df.set_index('净值日期')['累计单位净值']
        # 增加对比: 西部利得量化成长
        fund_name = '西部利得量化成长'
        sql_script = "SELECT a.jjdm fund_id, b.jzrq tradeDate, b.hbcl accumulate_return from " \
                     "funddb.jjxx1 a, funddb.jjhb b where a.cpfl = '2' and a.jjdm = b.jjdm " \
                     "and a.jjzt not in ('3', 'c') " \
                     "and a.m_opt_type <> '03' and a.jjdm = '{}' and b.jzrq >= {} and b.jzrq <= {} " \
                     "order by b.jzrq".format('000006', self.start_date, self.end_date)
        res = hbs.db_data_query('readonly', sql_script)
        data = pd.DataFrame(res['data'])
        data['ADJ_NAV'] = 0.01 * data['ACCUMULATE_RETURN'] + 1
        fund_nav_dict[fund_name] = data.set_index('TRADEDATE')['ADJ_NAV']

        # 周度版本
        if self.nav_frequency == 'week':
            trading_day_list = self._load_calendar()
            # 周度采样
            for name, nav_df in fund_nav_dict.items():
                fund_nav_dict[name] = nav_df.reindex(trading_day_list).dropna()
            # 补充几只
            f_dict = {"灵均进取1号": "SW3470", "鸣石春天十三号": "SX2090", "世纪前沿指数增强2号": "SGP682", "念空瑞景1号": "SGZ812"}
            for name, fund_id in f_dict.items():
                sql_script = "SELECT a.jjdm fund_id, b.jzrq TRADEDATE, b.fqdwjz as ADJ_NAV from " \
                             "st_hedge.t_st_jjxx a, st_hedge.t_st_rhb b where a.cpfl = '4' and a.jjdm = b.jjdm " \
                             "and a.jjzt not in ('3') " \
                             "and a.jjdm = '{}' and b.jzrq >= {} and b.jzrq <= {} " \
                             "order by b.jzrq".format(fund_id, self.start_date, self.end_date)
                res = hbs.db_data_query("highuser", sql_script, page_size=5000)
                data = pd.DataFrame(res['data'])
                fund_nav_dict[name] = data.set_index('TRADEDATE')['ADJ_NAV']
            # 邮件中提取的净值数据
            append_nav_dict = pd.read_excel(os.path.join(self.path, '佳期指增产品历史业绩.xlsx'), sheet_name=None)
            for fund_name, nav_df in append_nav_dict.items():
                nav_df['净值日期'] = nav_df['净值日期'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
                nav_df = nav_df[(nav_df['净值日期'] >= self.start_date) & (nav_df['净值日期'] <= self.end_date)]
                fund_nav_dict[fund_name] = nav_df.set_index('净值日期')['单位净值']

        self.fund_nav_dict = fund_nav_dict

    def _load_factor_data(self):
        """
        芯片：国证芯片
        锂电池：锂电池（中信）
        化工：细分化工
        有色：中证有色
        光伏：光伏产业
        @return:
        """
        factor_name_dict = {
            "创成长": "399296", "科创50": "000688", "有色": "930708", "化工": "000813",
            "芯片": "980017", "光伏": "931151", "新能源车": "399417"}
        factor_data = []
        for key, value in factor_name_dict.items():
            sql_script = "SELECT JYRQ as TRADEDATE, SPJG as TCLOSE from funddb.ZSJY WHERE ZQDM = '{}' and JYRQ >= " \
                         "{} and JYRQ <= {}".format(value, self.start_date, self.end_date)
            res = hbs.db_data_query('readonly', sql_script)
            data = pd.DataFrame(res['data'])
            # data['factor_ret'] = data['TCLOSE'].pct_change()
            data['factor_name'] = key
            data = data[['TRADEDATE', 'factor_name', 'TCLOSE']].dropna()
            factor_data.append(data)

        factor_data = pd.concat(factor_data)
        factor_data = pd.pivot_table(
            factor_data, index='TRADEDATE', columns='factor_name', values='TCLOSE').sort_index()

        self.factor_data = factor_data

    def _load_data(self):
        self._load_fund_nav_data()
        self._load_factor_data()

    @staticmethod
    def _perf_attr(account_return, factor_df, beta_series):
        calc_df = factor_df[beta_series.index.tolist()]

        idx = account_return.index.intersection(calc_df.index)
        calc_df = calc_df.loc[idx]

        factor_cum_return = calc_df.fillna(0.).add(1.0).prod().subtract(1.0)
        perf_attr = beta_series.multiply(factor_cum_return)
        perf_attr = pd.merge(beta_series.to_frame('factor_exposure'), perf_attr.to_frame('return_attr'),
                             left_index=True, right_index=True)

        return perf_attr

    def get_construct_result(self):
        # 回归参数
        lower, upper = None, None
        strategy_method, attribution_effort = 'ols', 'weak'

        attr_df = []
        for name, nav_df in self.fund_nav_dict.items():
            nav_df = nav_df.sort_index()
            idx = nav_df.index.intersection(self.factor_data.index)
            portfolio_return_series = nav_df.reindex(idx).pct_change().dropna()
            factor_df = self.factor_data.reindex(idx).pct_change().dropna()

            # from statsmodels.stats.outliers_influence import variance_inflation_factor
            # import numpy as np
            # name = factor_df.columns
            # x = np.matrix(factor_df)
            # VIF_list = [variance_inflation_factor(x, i) for i in range(x.shape[1])]
            # VIF = pd.DataFrame({'feature': name, "VIF": VIF_list})

            # factor_df = factor_df[['光伏', '化工', '有色', '芯片', '锂电池']]

            sr_obj = Regression(factor_df, portfolio_return_series,
                                upper=upper, lower=lower,
                                method=strategy_method, effort=attribution_effort)
            solve_dict = sr_obj.solve()
            factor_beta_series = solve_dict['solution_with_alpha']
            factor_beta_series.loc['r_square'] = solve_dict['r_square']

            factor_beta_series = factor_beta_series.to_frame('exposure')
            factor_beta_series.index.name = 'factor_name'
            factor_beta_series.reset_index(inplace=True)
            factor_beta_series['fund_name'] = name
            attr_df.append(factor_beta_series)

        attr_df = pd.concat(attr_df)
        attr_df = pd.pivot_table(attr_df, index='fund_name', columns='factor_name', values='exposure')
        cols = [x for x in attr_df.columns if x not in ['alpha', 'r_square']]
        if self.nav_frequency == 'day':
            attr_df['alpha'] *= 5
        attr_df.rename(columns={"alpha": "weekly_alpha", "fund_name": "标的名称"}, inplace=True)
        attr_df = attr_df[['weekly_alpha'] + cols + ['r_square']].round(5).reset_index()

        return attr_df


if __name__ == '__main__':
    attr_res = PrivateFundNavAttr('D:\\kevin\\绩效归因\\指增净值归因', '20201231', '20210806', 'day').get_construct_result()
    attr_res.to_csv('D:\\kevin\\attr_df.csv', index=False, encoding='gbk')



# import json
# import os

# class Config:
#
#     def __init__(self):
#
#         config_dir='//'.join(os.path.abspath(__file__).split('\\')[0:-1])+'//config'
#
#         with open(config_dir+"//Database_Connnection.json", "r") as load_f:
#             self.DBconfig= json.load(load_f)
#         with open(config_dir+"//Columns_Translation.json", "r") as load_f:
#             self.Columns_name_trans =  json.load(load_f)
#             # json.dump(new_dict, f)
#         with open(config_dir+"//JY_Sql.json", "r") as load_f:
#             self.Jy_sql =  json.load(load_f)
#         with open(config_dir+"//Mkr_code.json", "r") as load_f:
#             self.Mkr_code =  json.load(load_f)
#         with open(config_dir+"//IndexType.json", "r") as load_f:
#             self.Index_type= json.load(load_f)

class Config:

    def __init__(self):

        self.DBconfig={'Sql_ip': '192.168.223.152',
                     'Sql_user': 'admin',
                     'Sql_pass': 'mysql',
                     'Database': 'fe_temp_data',
                     'port': '3306'}

        self.Columns_name_trans={'科目代码': 'Code',
                                '科目名称': 'Name',
                                '数量': 'Quant',
                                '单位成本': 'Unit_cost',
                                '成本': 'Cost',
                                '成本占净值%': 'Cost_weight',
                                '市价': 'Price',
                                '市值': 'Mkt_value',
                                '市值占净值%': 'Weight',
                                '估值增值': 'Increased_val',
                                '停牌信息': 'Trading_flg',
                                '币种': 'Currency',
                                '汇率': 'Ex_rate',
                                '权益信息': 'Hd_info'}

        self.Jy_sql = {'sql_indus_a':
                           {'comment': 'join the industry of normal A share with its sec code and name table ',
                             'sql': 'SELECT A.CompanyCode,A.SecuCode,A.SecuAbbr,B.FirstIndustryName,B.XGRQ from HSJY_GG.SecuMain A left join HSJY_GG.LC_ExgIndustry B on A.CompanyCode=B.CompanyCode where B.Standard=24 @@'
                            },
                     'sql_indus_b':
                         {'comment': 'join the industry of  A share in chuangyeban  with its sec code and name table ',
                          'sql': 'SELECT A.CompanyCode,A.SecuCode,A.SecuAbbr,B.FirstIndustryName,B.UpdateTime from HSJY_GG.SecuMain A,HSJY_GP.LC_STIBExgIndustry B where A.CompanyCode=B.CompanyCode and B.Standard=24 @@'
                          },
                     'sql_fin':
                           {'comment': 'join the table LC_DIndicesForValuation and SecuMain to get the financial data',
                            'sql': 'select a.PE,a.PB,a.DividendRatio,a.TradingDay,b.SecuCode,b.SecuAbbr from HSJY_GG.LC_DIndicesForValuation a left join HSJY_GG.SecuMain b on a.InnerCode=b.InnerCode @@  '
                            },
                       'sql_fin_date':
                           {'comment': 'find the available date for in the LC_DIndicesForValuation table ',
                            'sql': "select distinct(TradingDay) from HSJY_GG.LC_DIndicesForValuation where TradingDay>= to_date('{0}','yyyymmdd') and TradingDay <=to_date('{1}','yyyymmdd')  "
                            },
                     'sql_benchmark':
                           {'comment': '',
                             'sql': 'select b.SecuCode,a.INNERCODE,a.WEIGHT,a.ENDDATE from hsjy_gg.LC_IndexComponentsWeight a,HSJY_GG.SecuMain b where a.INNERCODE=b.INNERCODE and IndexCode =@IndexCode @@  '
                            }
                       }

        self.Mkr_code ={'Ash': '11020101',
                         'Asz': '11023101',
                         'cyb': '11024101',
                         'ggt': '11028101',
                         'sgt': '11028301'}

        self.Index_type= {'沪深300': '000300', '中证500': '000905', '中证1000': '000852'}

        self.Asset_type= {'活期存款': '100201',
                         '上交所A股': '11020101',
                         '深交所A股': '11023101',
                         '深交所创业板': '11024101',
                         '上交所科创板': '1102C101',
                         '上交所新股成本': '11020301',
                         '上交所增发成本': '11020401',
                         '上交所非公开发行新股成本': '11020601',
                         '上交所配股成本': '11020701',
                         '深交所新股成本': '11023301',
                         '深交所增发成本': '11023401',
                         '深交所非公开发行新股成本': '11023601',
                         '深交所配股成本': '11023701',
                         '港股通': '11028101',
                         '深港通': '11028301',
                         '债券': '其中债券投资:',
                         '基金': '其中基金投资:',
                         '权证': '其中权证投资:',
                         '其他': '其中其他投资:'}



if __name__ == '__main__':
    from db_cons import sql_write_path_work, sql_user_work
    from hbshare.quant.CChen.cta_factor.hsjy_to_local import (
        hsjy_fut_pro_info,
        hsjy_fut_wr,
        hsjy_fut_com_info,
        hsjy_fut_com,
        hsjy_fut_member,
    )
    from hbshare.quant.CChen.cta_factor.hsjy_func import hsjy_fut_index
    from hbshare.quant.CChen.cta_factor.factor_index import run
    from hbshare.quant.CChen.cta_factor.factor_func import index_gen, index_compose
    import hbshare as hbs

    hbs.set_token('830ca3e1998947f8a99ebb7f3e563623')

    hsjy_fut_pro_info(
        db_path=sql_write_path_work['daily'],
        sql_info=sql_user_work
    )
    hsjy_fut_wr(
        db_path=sql_write_path_work['daily'],
        sql_info=sql_user_work
    )
    hsjy_fut_com_info(
        db_path=sql_write_path_work['daily'],
        sql_info=sql_user_work
    )
    hsjy_fut_com(
        db_path=sql_write_path_work['daily'],
        sql_info=sql_user_work
    )
    hsjy_fut_member(
        db_path=sql_write_path_work['daily'],
        sql_info=sql_user_work
    )
    hsjy_fut_index(
        db_path=sql_write_path_work['daily'],
        sql_info=sql_user_work
    )
    run(
        sql_path=sql_write_path_work['daily'],
        sql_info=sql_user_work
    )
    index_gen(
        sql_path=sql_write_path_work['daily'],
        sql_info=sql_user_work
    )
    index_compose(
        sql_path=sql_write_path_work['daily'],
        sql_info=sql_user_work
    )



#!/usr/bin/python
#coding:utf-8

"""
@author: Meng.lv
@contact: meng.lv@howbuy.com
@software: PyCharm
@file: __init__.py.py
@time: 2020/4/22 13:39
"""


from hbshare.quant.CChen.load_data import *
from hbshare.quant.CChen.fund_perf import *
# from hbshare.CChen.output import *
from hbshare.quant.CChen.fut import *



# -*- coding: utf-8 -*-

from datetime import datetime
from hbshare.fe.xwq.analysis.orm.hbdb import HBDB
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
plt.rcParams['font.sans-serif'] = ['SimHei']
plt.rcParams['axes.unicode_minus'] = False


def get_date(start_date, end_date):
    calendar_df = HBDB().read_cal(start_date, end_date)
    calendar_df = calendar_df.rename(columns={'JYRQ': 'CALENDAR_DATE', 'SFJJ': 'IS_OPEN', 'SFZM': 'IS_WEEK_END', 'SFYM': 'IS_MONTH_END'})
    calendar_df['CALENDAR_DATE'] = calendar_df['CALENDAR_DATE'].astype(str)
    calendar_df = calendar_df.sort_values('CALENDAR_DATE')
    calendar_df['IS_OPEN'] = calendar_df['IS_OPEN'].astype(int).replace({0: 1, 1: 0})
    calendar_df['YEAR_MONTH'] = calendar_df['CALENDAR_DATE'].apply(lambda x: x[:6])
    calendar_df['MONTH'] = calendar_df['CALENDAR_DATE'].apply(lambda x: x[4:6])
    calendar_df['MONTH_DAY'] = calendar_df['CALENDAR_DATE'].apply(lambda x: x[4:])
    calendar_df = calendar_df[(calendar_df['CALENDAR_DATE'] >= start_date) & (calendar_df['CALENDAR_DATE'] <= end_date)]
    trade_df = calendar_df[calendar_df['IS_OPEN'] == 1].rename(columns={'CALENDAR_DATE': 'TRADE_DATE'})
    trade_df = trade_df[(trade_df['TRADE_DATE'] >= start_date) & (trade_df['TRADE_DATE'] <= end_date)]
    report_df = calendar_df.drop_duplicates('YEAR_MONTH', keep='last').rename(columns={'CALENDAR_DATE': 'REPORT_DATE'})
    report_df = report_df[report_df['MONTH_DAY'].isin(['0331', '0630', '0930', '1231'])]
    report_df = report_df[(report_df['REPORT_DATE'] >= start_date) & (report_df['REPORT_DATE'] <= end_date)]
    report_trade_df = calendar_df[calendar_df['IS_OPEN'] == 1].rename(columns={'CALENDAR_DATE': 'TRADE_DATE'})
    report_trade_df = report_trade_df.sort_values('TRADE_DATE').drop_duplicates('YEAR_MONTH', keep='last')
    report_trade_df = report_trade_df[report_trade_df['MONTH'].isin(['03', '06', '09', '12'])]
    report_trade_df = report_trade_df[(report_trade_df['TRADE_DATE'] >= start_date) & (report_trade_df['TRADE_DATE'] <= end_date)]
    calendar_trade_df = calendar_df[['CALENDAR_DATE']].merge(trade_df[['TRADE_DATE']], left_on=['CALENDAR_DATE'], right_on=['TRADE_DATE'], how='left')
    calendar_trade_df['TRADE_DATE'] = calendar_trade_df['TRADE_DATE'].fillna(method='ffill')
    calendar_trade_df = calendar_trade_df[(calendar_trade_df['TRADE_DATE'] >= start_date) & (calendar_trade_df['TRADE_DATE'] <= end_date)]
    return calendar_df, report_df, trade_df, report_trade_df, calendar_trade_df

def get_cx_index(type, style, start_date, end_date):
    # 日历数据
    calendar_df, report_df, trade_df, report_trade_df, calendar_trade_df = get_date(start_date, end_date)
    # 个股市值数据
    stock_market_value = pd.read_hdf('D:/Git/hbshare/hbshare/fe/xwq/data/style_index/stock_market_value.hdf', key='table')
    # 个股行情数据
    stock_daily_k = pd.read_hdf('D:/Git/hbshare/hbshare/fe/xwq/data/style_index/stock_daily_k.hdf', key='table')
    stock_daily_k = stock_daily_k.pivot(index='TRADE_DATE', columns='TICKER_SYMBOL', values='CLOSE_PRICE').sort_index()
    stock_daily_k = stock_daily_k.replace(0.0, np.nan).fillna(method='ffill')
    stock_daily_ret = stock_daily_k.pct_change()
    stock_daily_ret = stock_daily_ret[(stock_daily_ret.index >= start_date) & (stock_daily_ret.index <= end_date)]
    stock_daily_ret = stock_daily_ret.unstack().reset_index()
    stock_daily_ret.columns = ['TICKER_SYMBOL', 'TRADE_DATE', 'RET']

    # 晨星股票风格
    stock_style = pd.read_hdf('D:/Git/hbshare/hbshare/fe/xwq/data/style_index/stock_style.hdf', key='table')
    stock_style = stock_style[stock_style['type'] == type]
    stock_style = stock_style[stock_style['category'] == style + '型']
    stock_style = stock_style[(stock_style['trade_date'] >= start_date) & (stock_style['trade_date'] <= end_date)]
    stock_style = stock_style.merge(calendar_trade_df.rename(columns={'CALENDAR_DATE': 'trade_date'}), on=['trade_date'], how='left')
    stock_style = stock_style.rename(columns={'ticker': 'TICKER_SYMBOL'}).merge(stock_market_value[['TRADE_DATE', 'TICKER_SYMBOL', 'MARKET_VALUE']], on=['TRADE_DATE', 'TICKER_SYMBOL'], how='left')
    total_mv = stock_style[['TRADE_DATE', 'MARKET_VALUE']].groupby('TRADE_DATE').sum().reset_index().rename(columns={'MARKET_VALUE': 'TOTAL_MARKET_VALUE'})
    stock_style = stock_style.merge(total_mv, on=['TRADE_DATE'], how='left')
    stock_style['WEIGHT'] = stock_style['MARKET_VALUE'] / stock_style['TOTAL_MARKET_VALUE']
    stock_style = stock_style.pivot(index='TRADE_DATE', columns='TICKER_SYMBOL', values='WEIGHT').sort_index().fillna(0.0)
    stock_style = trade_df[['TRADE_DATE']].set_index('TRADE_DATE').merge(stock_style, left_index=True, right_index=True, how='left')
    stock_style = stock_style.fillna(method='ffill')
    stock_style = stock_style.unstack().reset_index()
    stock_style.columns = ['TICKER_SYMBOL', 'TRADE_DATE', 'WEIGHT']
    stock_style = stock_style[stock_style['WEIGHT'] != 0.0]

    index = stock_style.merge(stock_daily_ret, on=['TICKER_SYMBOL', 'TRADE_DATE'], how='left')
    index['WEIGHT_RET'] = index['RET'] * index['WEIGHT']
    index = index[['TRADE_DATE', 'WEIGHT_RET']].groupby('TRADE_DATE').sum().reset_index().rename(columns={'WEIGHT_RET': style})
    index[style].iloc[0] = 0.0
    index[style] = (index[style] + 1.0).cumprod()
    index = index.set_index('TRADE_DATE')
    return index

if __name__ == '__main__':
    start_date = '20151231'
    end_date = '20220721'
    type = 'main'
    style_list = ['大盘成长', '大盘价值', '中盘成长', '中盘价值', '小盘成长', '小盘价值']
    for style in style_list:
        print(style)
        # 巨潮指数数据
        jc_df = HBDB().read_index_daily_k_given_date_and_indexs(start_date, ['399372', '399373', '399374', '399375', '399376', '399377'])
        jc_df = jc_df.pivot(index='jyrq', columns='zqmc', values='spjg')
        jc_df = jc_df[[style]]
        jc_df = jc_df / jc_df.iloc[0]
        jc_df = jc_df.rename(columns={style: style + '_巨潮'})
        jc_df.index = map(lambda x: datetime.strptime(str(x), '%Y%m%d'), jc_df.index)
        # 晨星指数数据
        cx_df = get_cx_index(type, style, start_date, end_date)
        cx_df = cx_df.rename(columns={style: style + '_晨星'})
        cx_df.index = map(lambda x: datetime.strptime(x, '%Y%m%d'), cx_df.index )

        plt.figure(figsize=(6, 6))
        plt.plot(jc_df.index, jc_df[style + '_巨潮'].values, label=style + '_巨潮', color='#F04950')
        plt.plot(cx_df.index, cx_df[style + '_晨星'].values, label=style + '_晨星', color='#6268A2')
        plt.legend(loc=2)
        plt.xticks(rotation=90)
        plt.tight_layout()
        plt.savefig('D:/Git/hbshare/hbshare/fe/xwq/data/style_index/{0}_{1}.png'.format(style, start_date))

    # # 巨潮相关
    # jc_daily_k = HBDB().read_index_daily_k_given_date_and_indexs('20220630', ['399311'])
    # jc_daily_k = jc_daily_k[jc_daily_k['jyrq'] == 20220630]
    # jc_cons = pd.read_hdf('D:/Git/hbshare/hbshare/fe/xwq/data/style_index/jc_cons.hdf', key='table')
    # jc_cons = jc_cons[jc_cons['ENDDATE'] == '2022-06-30 00:00:00']
    # jc_cons = jc_cons[['SECUCODE']].rename(columns={'SECUCODE': 'TICKER_SYMBOL'})
    # stock_mv = pd.read_hdf('D:/Git/hbshare/hbshare/fe/xwq/data/style_index/stock_market_value.hdf', key='table')
    # stock_mv = stock_mv[stock_mv['TRADE_DATE'] == '20220630']
    # stock_mv = jc_cons.merge(stock_mv[['TICKER_SYMBOL', 'MARKET_VALUE']], on=['TICKER_SYMBOL'], how='left')
    # stock_mv = stock_mv.sort_values('MARKET_VALUE', ascending=False).reset_index().drop('index', axis=1)
    # stock_mv['SIZE'] = np.nan
    # stock_mv['SIZE'].iloc[:200] = '大盘'
    # stock_mv['SIZE'].iloc[200:500] = '中盘'
    # stock_mv['SIZE'].iloc[500:] = '小盘'
    # stock_turnover = pd.read_excel('D:/Git/hbshare/hbshare/fe/xwq/data/style_index/A股成交额.xlsx')
    # stock_turnover['TICKER_SYMBOL'] = stock_turnover['TICKER_SYMBOL'].apply(lambda x: str(x).split('.')[0])
    # stock_mv = stock_mv.merge(stock_turnover, on=['TICKER_SYMBOL'], how='left')
    # stock_mv = stock_mv.dropna(subset=['TURNOVER_VALUE'])
    #
    # # 晨星相关
    # stock_mv = pd.read_hdf('D:/Git/hbshare/hbshare/fe/xwq/data/style_index/stock_market_value.hdf', key='table')
    # stock_mv = stock_mv[stock_mv['TRADE_DATE'] == '20220630']
    # stock_mv_total = stock_mv['MARKET_VALUE'].sum()
    # stock_mv_70 = stock_mv_total * 0.7
    # stock_mv_90 = stock_mv_total * 0.9
    # stock_mv = stock_mv.sort_values('MARKET_VALUE', ascending=False)
    # stock_mv['ACCUM_MARKET_VALUE'] = stock_mv['MARKET_VALUE'].cumsum()
    # stock_mv['SIZE'] = stock_mv['ACCUM_MARKET_VALUE'].apply(lambda x: '大盘' if x <= stock_mv_70 else '中盘' if x > stock_mv_70 and x <= stock_mv_90 else '小盘')
    # stock_turnover = pd.read_excel('D:/Git/hbshare/hbshare/fe/xwq/data/style_index/A股成交额.xlsx')
    # stock_turnover['TICKER_SYMBOL'] = stock_turnover['TICKER_SYMBOL'].apply(lambda x: str(x).split('.')[0])
    # stock_mv = stock_mv.merge(stock_turnover, on=['TICKER_SYMBOL'], how='left')
    # stock_mv = stock_mv.dropna(subset=['TURNOVER_VALUE'])